__all__ = ('HybridValueDictionary', )

from .dict_iterator_bases import DictionaryItemIteratorBase, DictionaryKeyIteratorBase, DictionaryValueIteratorBase
from .dict_update_iterable_iterator import _dict_update_iterable_iterator
from .docs import copy_docs, has_docs
from .weak_core import KeyedReferer, add_to_pending_removals, is_weakreferable
from .weak_value_dictionary import _WeakValueDictionaryCallback


class _HybridValueDictionaryKeyIterator(DictionaryKeyIteratorBase):
    __slots__ = ()
    
    
    @copy_docs(DictionaryKeyIteratorBase.__iter__)
    def __iter__(self):
        parent = self._parent
        parent._iterating += 1
        
        try:
            for key, (value_weakreferable, value_or_reference) in dict.items(parent):
                if value_weakreferable and (value_or_reference() is None):
                    add_to_pending_removals(parent, value_or_reference)
                    continue
                
                yield key
                continue
        
        finally:
            parent._iterating -= 1
            parent._commit_removals()
    
    
    @copy_docs(DictionaryKeyIteratorBase.__contains__)
    def __contains__(self, contains_key):
        return (contains_key in self._parent)


class _HybridValueDictionaryValueIterator(DictionaryValueIteratorBase):
    __slots__ = ()
    
    @copy_docs(DictionaryValueIteratorBase.__iter__)
    def __iter__(self):
        parent = self._parent
        parent._iterating += 1
        
        try:
            for value_weakreferable, value_or_reference in dict.values(parent):
                if value_weakreferable:
                    value = value_or_reference()
                    if value is None:
                        add_to_pending_removals(parent, value_or_reference)
                        continue
                else:
                    value = value_or_reference
                
                yield value
                continue
        
        finally:
            parent._iterating -=1
            parent._commit_removals()
    
    
    @copy_docs(DictionaryValueIteratorBase.__contains__)
    def __contains__(self, contains_value):
        parent = self._parent
        for value_weakreferable, value_or_reference in dict.values(parent):
            if value_weakreferable:
                value = value_or_reference()
                if value is None:
                    add_to_pending_removals(parent, value_or_reference)
                    continue
            else:
                value = value_or_reference
            
            if value == contains_value:
                result = True
                break
        else:
            result = False
        
        parent._commit_removals()
        
        return result


class _HybridValueDictionaryItemIterator(DictionaryItemIteratorBase):
    __slots__ = ()
    
    @copy_docs(DictionaryItemIteratorBase.__iter__)
    def __iter__(self):
        parent = self._parent
        parent._iterating += 1
        
        try:
            for key, (value_weakreferable, value_or_reference) in dict.items(parent):
                if value_weakreferable:
                    value = value_or_reference()
                    if value is None:
                        add_to_pending_removals(parent, value_or_reference)
                        continue
                else:
                    value = value_or_reference
                
                yield key, value
                continue
        
        finally:
            parent._iterating -= 1
            parent._commit_removals()
    
    
    @copy_docs(DictionaryItemIteratorBase.__contains__)
    def __contains__(self, contains_item):
        if not isinstance(contains_item, tuple):
            return False
        
        if len(contains_item) != 2:
            return False
        
        parent = self._parent
        contains_key, contains_value = contains_item
        
        value_pair = dict.get(parent, contains_key)
        if value_pair is None:
            return False
        
        value_weakreferable, value_or_reference = value_pair
        if value_weakreferable:
            value = value_or_reference()
            if value is None:
                if parent._iterating:
                    add_to_pending_removals(parent, value_or_reference)
                else:
                    dict.__delitem__(parent, contains_key)
                
                return False
        else:
            value = value_or_reference
        
        return (value == contains_value)


@has_docs
class HybridValueDictionary(dict):
    """
    Hybrid value dictionaries store their's values weakly referenced if applicable.
    
    Attributes
    ----------
    _pending_removals : `None`, `set` of (``KeyedReferer``, ``WeakHasher``)
        Pending removals of the hybrid value dictionary if applicable.
    _iterating : `int`
        Whether the hybrid value dictionary is iterating and how much times.
    _callback : ``_WeakValueDictionaryCallback``
        Callback added to the ``HybridValueDictionary``'s weak elements.
    
    Class Attributes
    ----------------
    MAX_REPR_ELEMENT_LIMIT : `int` = `50`
        The maximal amount of items to render by ``.__repr__``.
    
    Notes
    -----
    ``HybridValueDictionary``-s are weakreferable.
    """
    __slots__ = ('__weakref__', '_pending_removals', '_iterating', '_callback')
    
    MAX_REPR_ELEMENT_LIMIT = 50
    
    @has_docs
    def _commit_removals(self):
        """
        Commits the pending removals of the hybrid value dictionary if applicable.
        """
        if self._iterating:
            return
        
        pending_removals = self._pending_removals
        if pending_removals is None:
            return
        
        self._pending_removals = None
        
        for value_reference in pending_removals:
            key = value_reference.key
            try:
                value_pair = dict.__getitem__(self, key)
            except KeyError:
                continue
            
            value_weakreferable, value_or_reference = value_pair
            if (not value_weakreferable):
                continue
            
            if (value_or_reference is not value_reference):
                continue
            
            try:
                dict.__delitem__(self, key)
            except KeyError:
                pass
    
    # __class__ -> same
    
    @has_docs
    def __contains__(self, contains_key):
        """Returns whether the hybrid value dictionary contains the given key."""
        value_pair = dict.get(self, contains_key, None)
        if value_pair is None:
            return False
        
        value_weakreferable, value_or_reference = value_pair
        
        if value_weakreferable:
            if value_or_reference() is None:
                if self._iterating:
                    add_to_pending_removals(self, value_or_reference)
                else:
                    dict.__delitem__(self, contains_key)
                
                return False
        
        return True
    
    # __delattr__ -> same
    # __delitem__ -> same
    # __dir__ -> same
    # __doc__ -> same
    
    def __eq__(self, other):
        """Returns whether the two dictionaries are the same."""
        if isinstance(other, type(self)):
            return dict.__eq__(self, other)
        
        if isinstance(other, dict):
            pass
        
        elif hasattr(type(other), '__iter__'):
            try:
                other = dict(other)
            except (TypeError, ValueError):
                return NotImplemented
            
        else:
            return NotImplemented
        
        self_dict = dict(self.items())
        
        return self_dict == other
    
    # __format__ -> same
    # __ge__ -> same
    # __getattribute__ -> same
    
    @has_docs
    def __getitem__(self, key):
        """Gets the value of the hybrid value dictionary which matches the given key."""
        value_weakreferable, value_or_reference = dict.__getitem__(self, key)
        if value_weakreferable:
            value = value_or_reference()
            if value is None:
                if self._iterating:
                    add_to_pending_removals(self, value_or_reference)
                else:
                    dict.__delitem__(self, key)
                
                raise KeyError(key)
        else:
            value = value_or_reference
        
        return value
    
    # __gt__ -> same
    # __hash__ -> same
    
    @has_docs
    def __init__(self, iterable=None):
        """
        Creates a new ``HybridValueDictionary`` from the given iterable.
        
        Parameters
        ----------
        iterable : `None`, `iterable` = `None`, Optional
            Iterable to update the created dictionary with.
        """
        self._pending_removals = None
        self._iterating = 0
        self._callback = _WeakValueDictionaryCallback(self)
        if (iterable is not None):
            self.update(iterable)
    
    # __init_subclass__ -> same
    
    @has_docs
    def __iter__(self):
        """Returns a ``_HybridValueDictionaryKeyIterator`` iterating over the hybrid value dictionary's keys."""
        return iter(_HybridValueDictionaryKeyIterator(self))
    
    # __le__ -> same
    
    @has_docs
    def __len__(self):
        """Returns the length of the hybrid value dictionary."""
        length = dict.__len__(self)
        pending_removals = self._pending_removals
        if (pending_removals is not None):
            length -= len(pending_removals)
        
        return length
    
    # __lt__ -> same
    
    
    def __ne__(self, other):
        """Returns whether the two dictionaries are different."""
        if isinstance(other, type(self)):
            return dict.__ne__(self, other)
        
        if isinstance(other, dict):
            pass
        
        elif hasattr(type(other), '__iter__'):
            try:
                other = dict(other)
            except (TypeError, ValueError):
                return NotImplemented
        
        else:
            return NotImplemented
        
        self_dict = dict(self.items())
        
        return self_dict != other
    
    # __new__ -> same
    
    @has_docs
    def __reduce__(self):
        """Reduces the dictionary to a picklable object."""
        return (type(self), list(self.items()))
    
    
    @has_docs
    def __reduce_ex__(self, version):
        """Reduces the dictionary to a picklable object."""
        return type(self).__reduce__(self)
    
    
    @has_docs
    def __repr__(self):
        """Returns the representation of the hybrid value dictionary."""
        result = [self.__class__.__name__, '({']
        if len(self):
            limit = self.MAX_REPR_ELEMENT_LIMIT
            collected = 0
            for key, (value_weakreferable, value_or_reference) in dict.items(self):
                if value_weakreferable:
                    value = value_or_reference()
                    if value is None:
                        add_to_pending_removals(self, value_or_reference)
                        continue
                else:
                    value = value_or_reference
                
                result.append(repr(key))
                result.append(': ')
                result.append(repr(value))
                result.append(', ')
                
                collected += 1
                if collected != limit:
                    continue
                
                leftover = len(self) - collected
                if leftover:
                    result.append('...}, ')
                    result.append(str(leftover))
                    result.append(' truncated)')
                else:
                    result[-1] = '})'
                break
            else:
                result[-1] = '})'
            
            self._commit_removals()
        else:
            result.append('})')
        
        return ''.join(result)
    
    #__setattr__ -> same
    
    @has_docs
    def __setitem__(self, key, value):
        """Adds the given `key` - `value` pair to the hybrid value dictionary."""
        if is_weakreferable(value):
            value_weakreferable = True
            value_or_reference = KeyedReferer(value, self._callback, key)
        else:
            value_weakreferable = False
            value_or_reference = value
        
        dict.__setitem__(self, key, (value_weakreferable, value_or_reference), )
    
    # __sizeof__ -> same
    
    __str__ = __repr__
    
    # __subclasshook__ -> same
    
    @has_docs
    def clear(self):
        """
        Clears the hybrid value dictionary.
        """
        dict.clear(self)
        self._pending_removals = None
    
    @has_docs
    def copy(self):
        """
        Copies the hybrid value dictionary.
        
        Returns
        -------
        new : ``HybridValueDictionary``
        """
        new = dict.__new__(type(self))
        new._iterating = 0
        new._pending_removals = None
        callback = _WeakValueDictionaryCallback(new)
        new._callback = callback
        
        for key, (value_weakreferable, value_or_reference) in dict.items(self):
            if value_weakreferable:
                value = value_or_reference()
                if (value is None):
                    add_to_pending_removals(self, value_or_reference)
                    continue
                
                value_or_reference = KeyedReferer(value, callback, key)
            
            dict.__setitem__(new, key, (value_weakreferable, value_or_reference))
            continue
        
        self._commit_removals()
        
        return new
    
    @has_docs
    def get(self, key, default=None):
        """
        Gets the value of the hybrid value dictionary which matches the given key.
        
        Parameters
        ----------
        key : `Any`
            A key to match.
        default : `Any` = `None`, Optional
            Default value to return if the given `key` could not be matched.
        
        Returns
        -------
        value : `Any`, `default`
            The key's matched value. If no value was matched returns the `default` value.
        """
        value_pair = dict.get(self, key, default)
        if value_pair is default:
            return default
        
        value_weakreferable, value_or_reference = value_pair
        
        if value_weakreferable:
            value = value_or_reference()
            if value is None:
                if self._iterating:
                    add_to_pending_removals(self, value_or_reference)
                else:
                    dict.__delitem__(self, key)
                
                return default
        else:
            value = value_or_reference
        
        return value
    
    @has_docs
    def items(self):
        """
        Returns item iterator for the hybrid value dictionary.
        
        Returns
        -------
        item_iterator : ``_HybridValueDictionaryItemIterator``
        """
        return _HybridValueDictionaryItemIterator(self)
    
    @has_docs
    def keys(self):
        """
        Returns key iterator for the hybrid value dictionary.
        
        Returns
        -------
        key_iterator : ``_HybridValueDictionaryKeyIterator``
        """
        return _HybridValueDictionaryKeyIterator(self)
    
    # Need goto for better code-style
    @has_docs
    def pop(self, key, default=...):
        """
        Pops the value of the hybrid value dictionary which matches the given key.
        
        Parameters
        ----------
        key : `Any`
            A key to match.
        default : `Any` = `None`, Optional
            Default value to return if the given `key` could not be matched.
        
        Returns
        -------
        value : `Any`, `default`
            The key's matched value. If no value was matched and `default` value is given, then returns that.
        
        Raises
        ------
        KeyError
            If `key` could not be matched and `default` value is was not given either.
        """
        value_pair = dict.pop(self, key, ...)
        
        if (value_pair is not default):
            value_weakreferable, value_or_reference = value_pair
            
            if (not value_weakreferable):
                return value_or_reference
            
            value = value_or_reference()
            if (value is not None):
                return value
        
        if default is ...:
            raise KeyError(key)
        
        return default
    
    @has_docs
    def popitem(self):
        """
        Pops an item of the hybrid value dictionary.
        
        Returns
        -------
        item : `tuple` (`Any`, `Any`)
        
        Raises
        ------
        KeyError
            If the hybrid value dictionary is empty.
        """
        while dict.__len__(self):
            key, (value_weakreferable, value_or_reference) = dict.popitem(self)
            if value_weakreferable:
                value = value_or_reference()
                if value is None:
                    continue
            else:
                value = value_or_reference
                
            return key, value
        
        raise KeyError('popitem(): dictionary is empty.')
    
    @has_docs
    def setdefault(self, key, default=None):
        """
        Returns the value for the given `key`.
        
        If the `key` is not present in the hybrid value dictionary, then set's the given `default` value as it.
        
        Parameters
        ----------
        key : `Any`
            The key to match.
        default : `Any` = `None`, Optional
            Default value to set and return if `key` is not present in the hybrid value dictionary.
        
        Returns
        -------
        value : `default`, `Any`
            The matched value, or `default` if none.
        """
        value_pair = dict.get(self, key, None)
        if (value_pair is not None):
            value_weakreferable, value_or_reference = value_pair
            if (not value_weakreferable):
                return value_or_reference
            
            value = value_or_reference()
            if (value is not None):
                return value
        
        self[key] = default
        return default
    
    @has_docs
    def update(self, iterable):
        """
        Updates the hybrid value dictionary with the given iterable's elements.
        
        Parameters
        ----------
        iterable : `iterable`
            Iterable to extend the hybrid value dictionary with.
            
            Can be given as an object, which:
            - supports `.items` iterator.
            - supports `.keys` and `.__getitem__`.
            - is `iterable` and each iteration returns a sequence with 2 elements.
        
        Raises
        ------
        TypeError
            The given `iterable` is not `iterable`.
        ValueError
            The the given `iterable` sot supports neither `.items` or (`.keys` and `.__getitem__`) and one of it's
            element's length is not `2`.
        """
        iterable_type = iterable.__class__
        if hasattr(iterable_type, 'items'):
            for key, value in iterable.items():
                self[key] = value
            return
        
        if hasattr(iterable_type, 'keys') and hasattr(iterable_type, '__getitem__'):
            for key in iterable.keys():
                value = iterable[key]
                self[key] = value
            return
        
        if hasattr(iterable_type, '__iter__'):
            for key, value in _dict_update_iterable_iterator(iterable):
                self[key] = value
            
            return
        
        raise TypeError(
            f'`iterable` can be `iterable`, got {iterable_type.__name__}; {iterable_type!r}'
        )
    
    @has_docs
    def values(self):
        """
        Returns value iterator for the hybrid value dictionary.
        
        Returns
        -------
        value_iterator : ``_HybridValueDictionaryValueIterator``
        """
        return _HybridValueDictionaryValueIterator(self)



__all__ = ('AnsiTextDecoration', 'create_ansi_format_code',)


def create_ansi_format_code(text_decoration=None, background_color=None, foreground_color=None):
    """
    Creates an ansi text format code for `ansi` codeblocks.
    
    > If no parameter is given, will generate a format reset code.
    
    Parameters
    ----------
    text_decoration : `None`, `int` = `None`, Optional
        Text decoration.
    background_color : `None`, `tuple` (`int`, `int`, `int`) = `None`, Optional
        background color.
    foreground_color : `None`, `tuple` (`int`, `int`, `int`) = `None`, Optional
        Foreground color.
    
    Returns
    -------
    format_code : `str`
    """
    format_code_parst = ['\033[']
    
    field_added = False
    
    if (text_decoration is not None):
        field_added = True
        
        format_code_parst.append(str(text_decoration))
    
    if (background_color is not None):
        if field_added:
            format_code_parst.append(';')
        else:
            field_added = True
        
        red, green, blue = background_color
        
        format_code_parst.append('48;2;')
        format_code_parst.append(str(red))
        format_code_parst.append(';')
        format_code_parst.append(str(green))
        format_code_parst.append(';')
        format_code_parst.append(str(blue))
    
    if (foreground_color is not None):
        if field_added:
            format_code_parst.append(';')
        else:
            field_added = True
        
        red, green, blue = foreground_color
        
        format_code_parst.append('38;2;')
        format_code_parst.append(str(red))
        format_code_parst.append(';')
        format_code_parst.append(str(green))
        format_code_parst.append(';')
        format_code_parst.append(str(blue))
    
    if not field_added:
        format_code_parst.append(str(AnsiTextDecoration.reset_all))
    
    format_code_parst.append('m')
    
    return ''.join(format_code_parst)


class AnsiTextDecoration:
    """
    Contains the possible text decoration values as type attributes.
    """
    reset_all = 0
    bold = 1
    dim = 2
    italic = 3
    underline = 4
    slow_blink = 5
    rapid_blink = 6
    reverse = 7
    hidden = 8
    crossed = 9



__all__ = ('methodize',)

from types import MethodType

from .docs import has_docs


@has_docs
class methodize:
    """
    Wraps a type to as a method, allowing instancing it with it's parent instance object passed by default.
    
    Attributes
    ----------
    klass : `type`
        The type to instance as a method.
    """
    __slots__ = ('klass',)
    
    @has_docs
    def __init__(self, klass):
        """
        Creates a new ``methodize`` with the given class.
        
        Parameters
        ----------
        klass : `type`
             The type to instance as a method.
        """
        self.klass = klass
    
    def __get__(self, obj, type_):
        klass = self.klass
        if obj is None:
            return klass
        
        return MethodType(klass, obj)
    
    def __set__(self, obj, value):
        raise AttributeError('can\'t set attribute')
    
    def __delete__(self, obj):
        raise AttributeError('can\'t delete attribute')



__all__ = ('DEFAULT_ANSI_HIGHLIGHTER',)

from . import token_types as HIGHLIGHT_TOKEN_TYPES
from .ansi import create_ansi_format_code
from .formatter_context import HighlightFormatterContext


DEFAULT_ANSI_HIGHLIGHTER = HighlightFormatterContext()

# Add code highlights

DEFAULT_ANSI_HIGHLIGHTER.set_highlight_ansi_code(
    HIGHLIGHT_TOKEN_TYPES.TOKEN_TYPE_NON_SPACE,
    create_ansi_format_code(foreground_color = (220, 255, 255)),
)
DEFAULT_ANSI_HIGHLIGHTER.set_highlight_ansi_code(
    HIGHLIGHT_TOKEN_TYPES.TOKEN_TYPE_SPECIAL_OPERATOR,
    create_ansi_format_code(foreground_color = (255, 199, 222)),
)
DEFAULT_ANSI_HIGHLIGHTER.set_highlight_ansi_code(
    HIGHLIGHT_TOKEN_TYPES.TOKEN_TYPE_IDENTIFIER_MAGIC,
    create_ansi_format_code(foreground_color = (122, 255, 244)),
)
DEFAULT_ANSI_HIGHLIGHTER.set_highlight_ansi_code(
    HIGHLIGHT_TOKEN_TYPES.TOKEN_TYPE_NUMERIC,
    create_ansi_format_code(foreground_color = (96, 0, 255)),
)
DEFAULT_ANSI_HIGHLIGHTER.set_highlight_ansi_code(
    HIGHLIGHT_TOKEN_TYPES.TOKEN_TYPE_STRING,
    create_ansi_format_code(foreground_color = (174, 230, 151)),
)
DEFAULT_ANSI_HIGHLIGHTER.set_highlight_ansi_code(
    HIGHLIGHT_TOKEN_TYPES.TOKEN_TYPE_IDENTIFIER_KEYWORD,
    create_ansi_format_code(foreground_color = (255, 109, 109)),
)
DEFAULT_ANSI_HIGHLIGHTER.set_highlight_ansi_code(
    HIGHLIGHT_TOKEN_TYPES.TOKEN_TYPE_IDENTIFIER_BUILTIN,
    create_ansi_format_code(foreground_color = (0, 170, 255)),
)
DEFAULT_ANSI_HIGHLIGHTER.set_highlight_ansi_code(
    HIGHLIGHT_TOKEN_TYPES.TOKEN_TYPE_IDENTIFIER_BUILTIN_EXCEPTION,
    create_ansi_format_code(foreground_color = (176, 88, 255)),
)
DEFAULT_ANSI_HIGHLIGHTER.set_highlight_ansi_code(
    HIGHLIGHT_TOKEN_TYPES.TOKEN_TYPE_SPECIAL_PUNCTUATION,
    create_ansi_format_code(foreground_color = (157, 157, 255)),
)
DEFAULT_ANSI_HIGHLIGHTER.set_highlight_ansi_code(
    HIGHLIGHT_TOKEN_TYPES.TOKEN_TYPE_SPECIAL_OPERATOR_ATTRIBUTE,
    create_ansi_format_code(foreground_color = (255, 255, 194)),
)
DEFAULT_ANSI_HIGHLIGHTER.set_highlight_ansi_code(
    HIGHLIGHT_TOKEN_TYPES.TOKEN_TYPE_IDENTIFIER_ATTRIBUTE,
    create_ansi_format_code(foreground_color = (255, 255, 194)),
)
DEFAULT_ANSI_HIGHLIGHTER.set_highlight_ansi_code(
    HIGHLIGHT_TOKEN_TYPES.TOKEN_TYPE_SPECIAL_CONSOLE_PREFIX,
    create_ansi_format_code(foreground_color = (255, 121, 25)),
)
DEFAULT_ANSI_HIGHLIGHTER.set_highlight_ansi_code(
    HIGHLIGHT_TOKEN_TYPES.TOKEN_TYPE_STRING_UNICODE_FORMAT_MARK,
    create_ansi_format_code(foreground_color = (249, 17, 17)),
)
DEFAULT_ANSI_HIGHLIGHTER.set_highlight_ansi_code(
    HIGHLIGHT_TOKEN_TYPES.TOKEN_TYPE_COMMENT,
    create_ansi_format_code(foreground_color = (255, 152, 44)),
)
DEFAULT_ANSI_HIGHLIGHTER.set_highlight_ansi_code(
    HIGHLIGHT_TOKEN_TYPES.TOKEN_TYPE_STRING_UNICODE_FORMAT_CODE,
    create_ansi_format_code(foreground_color = (255, 152, 44)),
)
DEFAULT_ANSI_HIGHLIGHTER.set_highlight_ansi_code(
    HIGHLIGHT_TOKEN_TYPES.TOKEN_TYPE_STRING_UNICODE_FORMAT_POSTFIX,
    create_ansi_format_code(foreground_color = (155, 185, 243)),
)

# Add trace highlights

DEFAULT_ANSI_HIGHLIGHTER.set_highlight_ansi_code(
    HIGHLIGHT_TOKEN_TYPES.TOKEN_TYPE_TRACE,
    create_ansi_format_code(foreground_color = (255, 220, 220)),
)
DEFAULT_ANSI_HIGHLIGHTER.set_highlight_ansi_code(
    HIGHLIGHT_TOKEN_TYPES.TOKEN_TYPE_TRACE_TITLE,
    create_ansi_format_code(foreground_color = (235, 52, 113)),
)
DEFAULT_ANSI_HIGHLIGHTER.set_highlight_ansi_code(
    HIGHLIGHT_TOKEN_TYPES.TOKEN_TYPE_TRACE_LOCATION_PATH,
    create_ansi_format_code(foreground_color = (174, 230, 151)),
)
DEFAULT_ANSI_HIGHLIGHTER.set_highlight_ansi_code(
    HIGHLIGHT_TOKEN_TYPES.TOKEN_TYPE_TRACE_LOCATION_LINE_NUMBER,
    create_ansi_format_code(foreground_color = (151, 174, 230)),
)
DEFAULT_ANSI_HIGHLIGHTER.set_highlight_ansi_code(
    HIGHLIGHT_TOKEN_TYPES.TOKEN_TYPE_TRACE_LOCATION_NAME,
    create_ansi_format_code(foreground_color = (151, 174, 230)),
)

# Add console highlights

DEFAULT_ANSI_HIGHLIGHTER.set_highlight_ansi_code(
    HIGHLIGHT_TOKEN_TYPES.TOKEN_TYPE_CONSOLE_BANNER_LOGO,
    create_ansi_format_code(foreground_color = (137, 17, 217)),
)

DEFAULT_ANSI_HIGHLIGHTER.set_highlight_ansi_code(
    HIGHLIGHT_TOKEN_TYPES.TOKEN_TYPE_CONSOLE_BANNER_DESCRIPTION,
    create_ansi_format_code(foreground_color = (207, 252, 239)),
)

DEFAULT_ANSI_HIGHLIGHTER.set_highlight_ansi_code(
    HIGHLIGHT_TOKEN_TYPES.TOKEN_TYPE_CONSOLE_BANNER_DESCRIPTION_AWAIT_UNAVAILABLE,
    create_ansi_format_code(foreground_color = (235, 52, 113)),
)
DEFAULT_ANSI_HIGHLIGHTER.set_highlight_ansi_code(
    HIGHLIGHT_TOKEN_TYPES.TOKEN_TYPE_CONSOLE_MARKER,
    create_ansi_format_code(foreground_color = (174, 230, 151)),
)



__all__ = ('WeakSet',)

from .docs import has_docs
from .utils import is_hashable, is_iterable
from .weak_core import WeakReferer, add_to_pending_removals, is_weakreferable


@has_docs
class _WeakSetCallback:
    """
    Callback used by ``WeakSet``-s.
    
    Attributes
    ----------
    _parent : ``WeakReferer`` to ``WeakSet``
        The parent weak set.
    """
    __slots__ = ('_parent', )
    
    @has_docs
    def __new__(cls, parent):
        """
        Creates a new ``_WeakKeyDictionaryCallback`` bound to the given ``WeakSet``.
        
        Parameters
        ----------
        parent : ``WeakSet``
            The parent weak set.
        """
        parent = WeakReferer(parent)
        self = object.__new__(cls)
        self._parent = parent
        return self
    
    
    @has_docs
    def __call__(self, reference):
        """
        Called when a key of the respective weak set is garbage collected.
        
        Parameters
        ----------
        reference : ``WeakReferer``
            Weakreference to the respective object.
        """
        parent = self._parent()
        if parent is None:
            return
        
        if parent._iterating:
            add_to_pending_removals(parent, reference)
        else:
            set.discard(parent, reference)


class WeakSet(set):
    """
    Weak set, which stores it's elements weakly referenced.
    
    Attributes
    ----------
    _pending_removals : `None`, `set` of ``WeakReferer``
        Pending removals of the weak set if applicable.
    _iterating : `int`
        Whether the weak set is iterating and how much times.
    _callback : ``_WeakKeyDictionaryCallback``
        Callback added to the ``WeakSet``'s weak keys.
    
    Class Attributes
    ----------------
    MAX_REPR_ELEMENT_LIMIT : `int` = `50`
        The maximal amount of items to render by ``.__repr__``.
    
    Notes
    -----
    ``WeakSet``-s are weakreferable.
    """
    __slots__ = ('_pending_removals', '_iterating', '_callback')
    
    MAX_REPR_ELEMENT_LIMIT = 50


    @has_docs
    def _commit_removals(self):
        """
        Commits the pending removals of the weak set if applicable.
        """
        if self._iterating:
            return
        
        pending_removals = self._pending_removals
        if pending_removals is None:
            return
        
        self._pending_removals = None
        
        for key_reference in pending_removals:
            try:
                set.discard(self, key_reference)
            except KeyError:
                return
    
    @has_docs
    def __and__(self, other):
        """Returns a new set with the elements common of self and other."""
        if isinstance(other, type(self)):
            new = type(self)()
            
            for element_reference in set.__iter__(other):
                if set.__contains__(self, element_reference):
                    element = element_reference()
                    if (element is not None):
                        new.add(element)
            
            return new
        
        if is_iterable(other):
            new = type(self)()
            
            for element in other:
                if element in self:
                    try:
                        new.add(element)
                    except TypeError:
                        return NotImplemented
            
            return new
        
        return NotImplemented
    
    
    # __class__ -> same
    
    
    @has_docs
    def __contains__(self, element):
        """Returns whether the weak set contains the given element."""
        try:
            element = WeakReferer(element)
        except TypeError:
            return False
        
        return set.__contains__(self, element)
    
    
    # __delattr__ -> same
    # __dir__ -> same
    # __doc__ -> same
    
    
    def __eq__(self, other):
        if isinstance(other, type(self)):
            return set.__eq__(self, other)
        
        if isinstance(other, set):
            pass
        
        elif is_iterable(other):
            try:
                other = set(other)
            except TypeError:
                return NotImplemented
        
        else:
            return NotImplemented
        
        self_set = set(iter(self))
        
        return self_set == other
    
    
    # __format__ -> same
    
    
    @has_docs
    def __ge__(self, other):
        """Returns whether every element in other is in the set."""
        if isinstance(other, type(self)):
            return set.__ge__(self, other)
        
        elif isinstance(other, set):
            pass
        
        elif is_iterable(other):
            try:
                other = set(other)
            except TypeError:
                return NotImplemented
        else:
            return NotImplemented
        
        return set(iter(self)) >= other
    
    
    # __getattribute__ -> same
    
    
    @has_docs
    def __gt__(self, other):
        """Returns whether the set is a proper superset of other."""
        if isinstance(other, type(self)):
            return set.__gt__(self, other)
        
        elif isinstance(other, set):
            pass
        
        elif is_iterable(other):
            try:
                other = set(other)
            except TypeError:
                return NotImplemented
        else:
            return NotImplemented
        
        return set(iter(self)) > other
    
    
    # __hash__ ->same
    
    
    @has_docs
    def __iand__(self, other):
        """Update the set, keeping only elements found in it and all others."""
        if isinstance(other, type(self)):
            other_set = set(iter(other))
        
        elif isinstance(other, set):
            other_set = other
        
        elif is_iterable(other):
            try:
                other_set = set(other)
            except TypeError:
                return NotImplemented
        else:
            return NotImplemented
        
        elements_to_remove = []
        
        for element in self:
            if element not in other_set:
                elements_to_remove.append(element)
        
        for element in elements_to_remove:
            self.discard(element)
        
        return self
    
    
    @has_docs
    def __init__(self, iterable=None):
        """
        Creates a new ``WeakSet`` from the given iterable.
        
        Parameters
        ----------
        iterable : `None`, `iterable` = `None`, Optional
            Iterable to update the created set with.
        """
        self._pending_removals = None
        self._iterating = 0
        self._callback = _WeakSetCallback(self)
        if (iterable is not None):
            self.update(iterable)
    
    
    # __init_subclass__ -> same
    
    
    @has_docs
    def __ior__(self, other):
        """Updates the set with the element of others"""
        if isinstance(other, type(self)):
            pass
        
        elif isinstance(other, (set, dict)):
            for element in other:
                if not is_weakreferable(element):
                    return NotImplemented
            
        elif isinstance(other, (tuple, list)):
            for element in other:
                if not is_weakreferable(element):
                    return NotImplemented
                
                if not is_hashable(element):
                    return NotImplemented
                    
        elif is_iterable(other):
            # Make sure, we have unique elements, so convert other to set
            other = list(other)
            
            for element in other:
                if not is_weakreferable(element):
                    return NotImplemented
                
                if not is_hashable(element):
                    return NotImplemented
        
        else:
            return NotImplemented
        
        for element in other:
            self.add(element)
        
        return self
    
    
    @has_docs
    def __isub__(self, other):
        """Update the set, removing elements found in others."""
        if isinstance(other, type(self)):
            for element_reference in set.__iter__(other):
                set.discard(self, element_reference)
            
            return self
        
        elif is_iterable(other):
            pass
        
        else:
            return NotImplemented
        
        for element in other:
            self.discard(element)
        
        return self
    
    
    @has_docs
    def __iter__(self):
        """
        Iterates over a weak set's elements.
        
        This method is a generator.
        
        Yields
        ------
        element : `Any`
        """
        self._iterating += 1
        
        try:
            for key_reference in set.__iter__(self):
                key = key_reference()
                if (key is None):
                    add_to_pending_removals(self, key_reference)
                    continue
                
                yield key
                continue
        
        finally:
            self._iterating -= 1
            self._commit_removals()
    
    
    @has_docs
    def __ixor__(self, other):
        """Update the set, keeping only elements found in either set, but not in both."""
        if isinstance(other, type(self)):
            for element_reference in set.__iter__(other):
                element = element_reference()
                if element is None:
                    add_to_pending_removals(self, element_reference)
                    continue
                
                try:
                    set.remove(self, element_reference)
                except KeyError:
                    self.add(element)
            
            return self
        
        if isinstance(other, (set, dict)):
            for element in other:
                if not is_weakreferable(element):
                    return NotImplemented
            
        elif is_iterable(other):
            # Make sure to check out only unique elements
            try:
                other = set(other)
            except TypeError:
                return NotImplemented
            
            for element in other:
                if not is_weakreferable(element):
                    return NotImplemented
        
        else:
            return NotImplemented
        
        for element in other:
            try:
                self.remove(element)
            except KeyError:
                self.add(element)
        
        return self
    
    
    @has_docs
    def __le__(self, other):
        """Returns whether every element in the set is in other."""
        if isinstance(other, type(self)):
            return set.__le__(self, other)
        
        elif isinstance(other, set):
            pass
        
        elif is_iterable(other):
            try:
                other = set(other)
            except TypeError:
                return NotImplemented
        else:
            return NotImplemented
        
        return set(iter(self)) <= other
    
    
    @has_docs
    def __len__(self):
        """Returns the length of the set."""
        length = set.__len__(self)
        pending_removals = self._pending_removals
        if (pending_removals is not None):
            length -= len(pending_removals)
        
        return length
    
    
    @has_docs
    def __lt__(self, other):
        """Test whether the set is a proper superset of other"""
        if isinstance(other, type(self)):
            return set.__lt__(self, other)
        
        elif isinstance(other, set):
            pass
        
        elif is_iterable(other):
            try:
                other = set(other)
            except TypeError:
                return NotImplemented
        else:
            return NotImplemented
        
        return set(iter(self)) < other
    
    
    def  __ne__(self, other):
        if isinstance(other, type(self)):
            return set.__ne__(self, other)
        
        if isinstance(other, set):
            pass
        
        elif is_iterable(other):
            try:
                other = set(other)
            except TypeError:
                return NotImplemented
        else:
            return NotImplemented
        
        self_elements = set(iter(self))
        
        return self_elements != other
    
    
    # __new__-> same
    
    
    @has_docs
    def __or__(self, other):
        """Returns the union of two sets."""
        if isinstance(other, type(self)):
            new = self.copy()
            
            for element in other:
                new.add(element)
        
        elif is_iterable(other):
            new = self.copy()
            
            for element in other:
                try:
                    new.add(element)
                except TypeError:
                    return NotImplemented
            
        else:
            return NotImplemented
        
        return new
    
    
    __rand__ = __and__
    
    
    @has_docs
    def __reduce__(self):
        """Reduces the set to a picklable object."""
        return (type(self), list(self))
    
    
    @has_docs
    def __reduce_ex__(self, version):
        """Reduces the set to a picklable object."""
        return type(self).__reduce__(self)
    
    
    @has_docs
    def __repr__(self):
        """Returns the set's representation."""
        result = [self.__class__.__name__, '({']
        if len(self):
            limit = self.MAX_REPR_ELEMENT_LIMIT
            collected = 0
            
            for element_reference in set.__iter__(self):
                element = element_reference()
                if (element is None):
                    add_to_pending_removals(self, element_reference)
                    continue
                
                result.append(repr(element))
                result.append(', ')
                
                collected +=1
                if collected != limit:
                    continue
                
                leftover = len(self) - collected
                if leftover:
                    result.append('...}, ')
                    result.append(str(leftover))
                    result.append(' truncated)')
                else:
                    result[-1] = '})'
                break
            else:
                result[-1] = '})'
            
            self._commit_removals()
        else:
            result.append('})')
        
        return ''.join(result)
    
    
    __ror__ = __or__
    
    
    @has_docs
    def __rsub__(self, other):
        """Returns a new set, without elements found in self."""
        if isinstance(other, type(self)):
            other = set(iter(other))
        
        elif isinstance(other, set):
            pass
        
        elif is_iterable(other):
            try:
                other = set(other)
            except TypeError:
                return NotImplemented
        
        else:
            return NotImplemented
        
        new = type(self)()
        
        for element in other:
            if element not in self:
                try:
                    new.add(element)
                except TypeError:
                    return NotImplemented
        
        return new
    
    
    @has_docs
    def __rxor__(self, other):
        """Return a new set with elements in either the set or other but not both."""
        if isinstance(other, type(self)):
            other = set(iter(other))
        
        elif isinstance(other, set):
            pass
        
        elif is_iterable(other):
            try:
                other = set(other)
            except TypeError:
                return NotImplemented
        
        else:
            return NotImplemented
        
        self_set = set(iter(self))
        
        new = type(self)()
        
        for element in (self_set ^ other):
            try:
                new.add(element)
            except TypeError:
                return NotImplemented
        
        return new
    
    
    # __setattr__ -> same
    # __sizeof__ -> same
    
    
    __str__ = __repr__
    
    
    @has_docs
    def __sub__(self, other):
        """Returns a new set, without elements found in other."""
        if isinstance(other, type(self)):
            other = set(iter(other))
        
        elif isinstance(other, set):
            pass
        
        elif is_iterable(other):
            try:
                other = set(other)
            except TypeError:
                return NotImplemented
        
        else:
            return NotImplemented
        
        new = type(self)()
        
        for element in self:
            if element not in other:
                new.add(element)
        
        return new
    
    # __subclasshook__ -> same
    
    
    __xor__ = __rxor__
    
    
    @has_docs
    def add(self, element):
        """
        Adds the `element` to the set.
        
        Parameters
        ----------
        element : `Any`
            The element to add.
        
        Raises
        ------
        TypeError
            If `element` is not weakreferable.
        """
        element_reference = WeakReferer(element, self._callback)
        set.add(self, element_reference)
    
    
    @has_docs
    def clear(self):
        """Remove all elements from the set."""
        self._pending_removals = None
        set.clear(self)
    
    
    @has_docs
    def copy(self):
        """Return a shallow copy of the set."""
        new = type(self)()
        new_callback = new._callback
        
        for element_reference in set.__iter__(self):
            element = element_reference()
            if (element is not None):
                element_reference =  WeakReferer(element, new_callback)
                set.add(new, element_reference)
        
        return new
    
    
    def difference(self, iterable):
        """
        Returns a new set, without elements found in other.
        
        Parameters
        ----------
        iterable : `iterable`
            The values to get difference with.
        
        Returns
        -------
        new : ``WeakSet``
        
        Raises
        ------
        TypeError
            - If `iterable` is not `iterable`.
            - If an element of `iterable` is not hashable.
            - If an element of `iterable` is now weakreferable.
        """
        if isinstance(iterable, type(self)):
            iterable = set(iter(iterable))
        
        elif isinstance(iterable, set):
            pass
        
        elif is_iterable(iterable):
            iterable = set(iterable)
        
        else:
            raise TypeError(
                f'`iterable` can be `iterable`, got {iterable.__class__.__name__}; {iterable!r}.'
            )
        
        new = type(self)()
        
        for element in self:
            if element not in iterable:
                new.add(element)
        
        return new
    
    
    def difference_update(self, iterable):
        """
        Update the set, removing elements found in others.
        
        Parameters
        ----------
        iterable : `iterable`
            The values to make difference with.
        
        Raises
        ------
        TypeError
            - If `iterable` is not `iterable`.
            - If an element of `iterable` is not hashable.
            - If an element of `iterable` is now weakreferable.
        """
        if isinstance(iterable, type(self)):
            for element_reference in set.__iter__(iterable):
                set.discard(self, element_reference)
        
        elif is_iterable(iterable):
            for element in iterable:
                self.discard(element)
        
        else:
            raise TypeError(
                f'`iterable` can be `iterable`, got {iterable.__class__.__name__}; {iterable!r}.'
            )
    
    
    @has_docs
    def discard(self, element):
        """
        Removes `element` from the set if it is present.
        
        Parameters
        ----------
        element : `Any`
            The element to remove.
        """
        try:
            element_reference = WeakReferer(element)
        except TypeError:
            pass
        else:
            set.discard(self, element_reference)
    
    
    @has_docs
    def intersection(self, iterable):
        """
        Returns a new set with the elements common of self and other.
        
        Parameters
        ----------
        iterable : `iterable`
            The values to make difference with.
        
        Returns
        -------
        new : ``WeakSet``
        
        Raises
        ------
        TypeError
            - If `iterable` is not `iterable`.
            - If an element of `iterable` is not hashable.
            - If an element of `iterable` is now weakreferable.
        """
        if isinstance(iterable, type(self)):
            new = type(self)()
            
            for element_reference in set.__iter__(iterable):
                if set.__contains__(self, element_reference):
                    element = element_reference()
                    if (element is not None):
                        new.add(element)
            
            return new
        
        if is_iterable(iterable):
            new = type(self)()
            
            for element in iterable:
                if element in self:
                    new.add(element)
            
            return new
        
        raise TypeError(
            f'`iterable` can be `iterable`, got {iterable.__class__.__name__}; {iterable!r}.'
        )
    
    
    @has_docs
    def intersection_update(self, iterable):
        """
        Update the set, keeping only elements found in it and all others.
        
        Parameters
        ----------
        iterable : `iterable`
            The values to make difference with.
        
        Raises
        ------
        TypeError
            - If `iterable` is not `iterable`.
            - If an element of `iterable` is not hashable.
        """
        if isinstance(iterable, type(self)):
            iterable_set = set(iter(iterable))
        
        elif isinstance(iterable, set):
            iterable_set = iterable
        
        elif is_iterable(iterable):
            iterable_set = set(iterable)
        
        else:
            raise TypeError(
                f'`iterable` can be `iterable`, got {iterable.__class__.__name__}; {iterable!r}.'
            )
        
        elements_to_remove = []
        
        for element in self:
            if element not in iterable_set:
                elements_to_remove.append(element)
        
        for element in elements_to_remove:
            self.discard(element)
    
    
    @has_docs
    def isdisjoint(self, iterable):
        """
        Return whether if the set has no elements in common with other. Sets are disjoint if and only if
        their intersection is the empty set.
        
        Parameters
        ----------
        iterable : `iterable`
            The value to check disjoint with.
        
        Returns
        -------
        is_disjoint : `bool`
        
        Raises
        ------
        TypeError
            - If `iterable` is not `iterable`.
            - If an element of `iterable` is not hashable.
        """
        if isinstance(iterable, type(self)):
            iterable = set(iter(iterable))
            
            # Something breaks the normal solution:
            # return set.isdisjoint(self, other)
        
        elif isinstance(iterable, set):
            pass
        
        elif is_iterable(iterable):
            iterable = set(iterable)
        
        else:
            raise TypeError(
                f'`iterable` can be `iterable`, got {iterable.__class__.__name__}; {iterable!r}.'
            )
        
        self_set = set(iter(self))
        return self_set.isdisjoint(iterable)
    
    
    @has_docs
    def issubset(self, iterable):
        """
        Returns whether every element in the set is in other.
        
        Parameters
        ----------
        iterable : `iterable`
            The value to check subset state with.
        
        Returns
        -------
        is_subset : `bool`
        
        Raises
        ------
        TypeError
            - If `iterable` is not `iterable`.
            - If an element of `iterable` is not hashable.
        """
        if isinstance(iterable, type(self)):
            return set.__le__(self, iterable)
        
        elif isinstance(iterable, set):
            pass
        
        elif is_iterable(iterable):
            iterable = set(iterable)
        
        else:
            raise TypeError(
                f'`iterable` can be `iterable`, got {iterable.__class__.__name__}; {iterable!r}.'
            )
        
        return set(iter(self)).issubset(iterable)
    
    
    @has_docs
    def issuperset(self, iterable):
        """
        Returns whether every element in other is in the set.
        
        Parameters
        ----------
        iterable : `iterable`
            The value to check superset state with.
        
        Returns
        -------
        is_superset : `bool`
        
        Raises
        ------
        TypeError
            - If `iterable` is not `iterable`.
            - If an element of `iterable` is not hashable.
        """
        if isinstance(iterable, type(self)):
            return set.__ge__(self, iterable)
        
        elif isinstance(iterable, set):
            pass
        
        elif is_iterable(iterable):
            iterable = set(iterable)
        
        else:
            raise TypeError(
                f'`iterable` can be `iterable`, got {iterable.__class__.__name__}; {iterable!r}.'
            )
        
        return set(iter(self)).issuperset(iterable)
    
    
    @has_docs
    def pop(self):
        """
        Remove and return an arbitrary element from the set.
        
        Returns
        -------
        element : `Any`
            The removed element.
        
        Raises
        ------
        KeyError
            If the set is empty.
        """
        while True:
            element_reference = set.pop(self)
            
            element = element_reference()
            if (element is not None):
                return element
    
    
    @has_docs
    def remove(self, element):
        """
        Removes `element` from the set.
        
        Parameters
        ----------
        element : `Any`
            The element ot remove.
        
        Raises
        ------
        KeyError
            If `element` is not in the set.
        """
        try:
            element_reference = WeakReferer(element)
        except TypeError:
            raise KeyError(element) from None
        
        set.remove(self, element_reference)
    
    
    @has_docs
    def symmetric_difference(self, iterable):
        """
        Return a new set with elements in either the set or other but not both.
        
        Parameters
        ----------
        iterable : `iterable`
            The values to make symmetric difference with.
        
        Returns
        -------
        new : ``WeakSet``
        
        Raises
        ------
        TypeError
            - If `iterable` is not `iterable`.
            - If an element of `iterable` is not hashable.
            - If an element of `iterable` is now weakreferable.
        """
        if isinstance(iterable, type(self)):
            iterable = set(iter(iterable))
        
        elif isinstance(iterable, set):
            pass
        
        elif is_iterable(iterable):
            iterable = set(iterable)
        
        else:
            raise TypeError(
                f'`iterable` can be `iterable`, got {iterable.__class__.__name__}; {iterable!r}.'
            )
        
        self_set = set(iter(self))
        
        new = type(self)()
        
        for element in (self_set ^ iterable):
            new.add(element)
        
        return new
    
    
    @has_docs
    def symmetric_difference_update(self, iterable):
        """
        Update the set, keeping only elements found in either set, but not in both.
        
        Parameters
        ----------
        iterable : `iterable`
            The values to do symmetric difference update with.
        
        Returns
        -------
        new : ``WeakSet``
        
        Raises
        ------
        TypeError
            - If `iterable` is not `iterable`.
            - If an element of `iterable` is not hashable.
            - If an element of `iterable` is now weakreferable.
        """
        if isinstance(iterable, type(self)):
            for element_reference in set.__iter__(iterable):
                element = element_reference()
                if element is None:
                    add_to_pending_removals(self, element_reference)
                    continue
                
                try:
                    set.remove(self, element_reference)
                except KeyError:
                    self.add(element)
            
            return
        
        
        if isinstance(iterable, (set, dict)):
            for element in iterable:
                try:
                    self.remove(element)
                except KeyError:
                    self.add(element)
        
        
        elif is_iterable(iterable):
            # check for dupes
            added = set()
            
            for element in iterable:
                if (element not in added):
                    try:
                        self.remove(element)
                    except KeyError:
                        self.add(element)
                    
                    added.add(element)
        
        else:
            raise TypeError(
                f'`iterable` can be `iterable`, got {iterable.__class__.__name__}; {iterable!r}.'
            )
    
    
    @has_docs
    def union(self, iterable):
        """
        Returns the union of two sets.
        
        Parameters
        ----------
        iterable : `iterable`
            The values to union with.
        
        Returns
        -------
        new : ``WeakSet``
        
        Raises
        ------
        TypeError
            - If `iterable` is not `iterable`.
            - If an element of `iterable` is not hashable.
            - If an element of `iterable` is now weakreferable.
        """
        if isinstance(iterable, type(self)) or is_iterable(iterable):
            new = self.copy()
            
            for element in iterable:
                new.add(element)
        
        else:
            raise TypeError(
                f'`iterable` can be `iterable`, got {iterable.__class__.__name__}; {iterable!r}.'
            )
        
        return new
    
    
    def update(self, iterable):
        """
        Updates the set in-place.
        
        Parameters
        ----------
        iterable : `iterable`
            Iterable to extend the set's elements with.
        
        Raises
        ------
        TypeError
            - If `iterable` is not iterable.
            - If an element of `iterable` is not hashable.
            - If an element of `iterable` is not weakreferable.
        """
        if isinstance(iterable, type(self)) or hasattr(iterable, '__iter__'):
            for element in iterable:
                self.add(element)
        
        else:
            raise TypeError(
                f'`iterable` can be `iterable`, got {iterable.__class__.__name__}; {iterable!r}.'
            )



from .highlight import *

from .analyzer import *
from .async_utils import *
from .base_method import *
from .code import *
from .compact import *
from .compound_type import *
from .dict_iterator_bases import *
from .dict_update_iterable_iterator import *
from .docs import *
from .export_include import *
from .frame import *
from .function_tools import *
from .hybrid_value_dictionary import *
from .ignore_case_multi_value_dictionary import *
from .ignore_case_string import *
from .json import *
from .keep_type import *
from .method_like import *
from .multi_value_dictionary import *
from .properties import *
from .removed_descriptor import *
from .rich_attribute_error import *
from .sorted_list import *
from .utils import *
from .trace import *
from .type_methodizer import *
from .type_modulizer import *
from .weak_core import *
from .weak_item_dictionary import *
from .weak_key_dictionary import *
from .weak_map import *
from .weak_set import *
from .weak_value_dictionary import *

__all__ = (
    *highlight.__all__,
    
    *analyzer.__all__,
    *async_utils.__all__,
    *base_method.__all__,
    *code.__all__,
    *compact.__all__,
    *compound_type.__all__,
    *dict_iterator_bases.__all__,
    *dict_update_iterable_iterator.__all__,
    *docs.__all__,
    *export_include.__all__,
    *frame.__all__,
    *function_tools.__all__,
    *hybrid_value_dictionary.__all__,
    *ignore_case_multi_value_dictionary.__all__,
    *ignore_case_string.__all__,
    *json.__all__,
    *keep_type.__all__,
    *method_like.__all__,
    *multi_value_dictionary.__all__,
    *properties.__all__,
    *removed_descriptor.__all__,
    *rich_attribute_error.__all__,
    *sorted_list.__all__,
    *utils.__all__,
    *trace.__all__,
    *type_methodizer.__all__,
    *type_modulizer.__all__,
    *weak_core.__all__,
    *weak_item_dictionary.__all__,
    *weak_key_dictionary.__all__,
    *weak_map.__all__,
    *weak_set.__all__,
    *weak_value_dictionary.__all__,
)



__all__ = ('create_event_loop', 'get_event_loop', 'set_event_loop')

from threading import current_thread, enumerate as list_threads

from ...utils import WeakItemDictionary, export

from ..event_loop import EventThread


THREAD_TO_EVENT_LOOP_REFERENCE = WeakItemDictionary()


def _try_detect_event_loop(local_thread):
    """
    Tries to detect the event loop, what should be used from the current thread.
    
    Returns
    -------
    event_loop : `None`, ``EventThread``
    """
    try:
        thread = THREAD_TO_EVENT_LOOP_REFERENCE[local_thread]
    except KeyError:
        pass
    else:
        if not thread.is_stopped():
            return thread
    
    for thread in list_threads():
        if not isinstance(thread, EventThread):
            continue
        
        if thread.daemon:
            continue
        
        if thread.is_stopped():
            continue
        
        THREAD_TO_EVENT_LOOP_REFERENCE[local_thread] = thread
        return thread


def _get_event_loop_is_current_thread(loop):
    """
    Gets the local event loop if applicable and whether it is indeed the current thread or nah.
    
    Used by other top level functions.
    
    Parameters
    ----------
    loop : `None`, ``EventLoop``
        Optional event loop to use instead of auto-detecting.
    
    Returns
    -------
    event_loop : ``EventThread``
        The local event loop.
    is_current_thread : `bool`
        Whether the event loop is the current thread.
    
    Raises
    ------
    RuntimeError
        There are are no detectable event loops.
    """
    local_thread = current_thread()
    if (loop is not None):
        return loop, (loop is local_thread)
    
    if isinstance(local_thread, EventThread):
        return local_thread, True
    
    event_loop = _try_detect_event_loop(local_thread)
    if (event_loop is not None):
        return event_loop, False
    
    raise RuntimeError(
        f'No running event loop for {local_thread!r}.'
    )


@export
def get_event_loop():
    """
    Gets the local event loop if applicable. If not raises `RuntimeError`.
    
    Returns
    -------
    event_loop : ``EventThread``
        The local event loop.
    
    Raises
    ------
    RuntimeError
        There are are no detectable event loops.
    """
    local_thread = current_thread()
    if isinstance(local_thread, EventThread):
        return local_thread
    
    event_loop = _try_detect_event_loop(local_thread)
    if (event_loop is not None):
        return event_loop
    
    raise RuntimeError(
        f'No running event loop for {local_thread!r}.'
    )


def set_event_loop(event_loop):
    """
    Sets the given event loop to be linked to the current thread.
    
    Can be used to force update event loop resolution of ``get_event_loop`` for the current thread.
    
    Parameters
    ----------
    event_loop : ``EventThread``
        The event loop to set.
    
    Raises
    ------
    RuntimeError
        Cannot set event loop from an event loop.
    """
    local_thread = current_thread()
    if isinstance(local_thread, EventThread):
        raise RuntimeError(
            f'Cannot set event loop from an event loop. Current event loop: {local_thread!r} ;got: {event_loop!r}.'
        )
    
    THREAD_TO_EVENT_LOOP_REFERENCE[local_thread] = event_loop


def _maybe_set_event_loop(event_loop):
    """
    Sets the event loop linked to the current thread if possible.
    
    Parameters
    ----------
    event_loop : ``EventThread``
        The event loop to set.
    """
    local_thread = current_thread()
    if isinstance(local_thread, EventThread):
        return
    
    linked_event_loop = _try_detect_event_loop(local_thread)
    if (linked_event_loop is not None):
        return
    
    THREAD_TO_EVENT_LOOP_REFERENCE[local_thread] = event_loop


def create_event_loop(**kwargs):
    """
    Creates a new event loop.
    
    Parameters
    ----------
    **kwargs : Keyword parameters
        Parameters to create the event loop with.
    
    Other parameters
    ----------------
    daemon : `bool` = `False`, Optional (Keyword only)
        Whether the event loop should be daemon.
    name : `None`, `str` = `None`, Optional (Keyword only)
        The event loop's name.
    start_later : `bool` = `True`, Optional (Keyword only)
        Whether the event loop should be started only later.

    Returns
    -------
    event_loop : ``EventThread``
    """
    event_loop = EventThread(**kwargs)
    _maybe_set_event_loop(event_loop)
    return event_loop



__all__ = ()

from ..docs import copy_docs
from ..export_include import export

from .matching import PYTHON_PARSERS, PYTHON_PARSERS_FORMAT_STRING, _try_match_till_format_string_expression
from .token import Token, _merge_tokens
from .token_types import (
    MERGE_TOKEN_TYPES, TOKEN_TYPE_ALL, TOKEN_TYPE_COMMENT, TOKEN_TYPE_LINEBREAK, TOKEN_TYPE_LINEBREAK_ESCAPED,
    TOKEN_TYPE_SPACE, TOKEN_TYPE_SPECIAL_OPERATOR, TOKEN_TYPE_SPECIAL_PUNCTUATION, TOKEN_TYPE_STRING_UNICODE_FORMAT,
    TOKEN_TYPE_STRING_UNICODE_FORMAT_CODE, TOKEN_TYPE_STRING_UNICODE_FORMAT_MARK
)


class HighlightParserContextBase:
    """
    Base class for highlighting.
    
    Attributes
    ----------
    done : `bool`
        Whether processing is done.
    tokens : `list` of ``Token``
        The generated tokens.
    """
    __slots__ = ('done', 'tokens',)
    
    def __new__(cls):
        """
        Creates a new ``HighlightParserContextBase``.
        """
        return object.__new__(cls)
    
    
    def get_line_index(self):
        """
        Returns the line's index where the context is at.
        
        Returns
        -------
        line_index : `int`
        """
        return 0
    
    
    def get_line(self):
        """
        Returns the actual line of the context.
        
        Returns
        -------
        line : `str`
        """
        return ''
    
    
    def get_line_character_index(self):
        """
        Returns the character index of the context's actual line.
        
        Returns
        -------
        line_character_index : `int`
        """
        return 0
    
    
    def set_line_character_index(self, line_character_index):
        """
        Sets the actual line's character index.
        
        Parameters
        ----------
        line_character_index : `int`
            The index to set of the actual line.
            
            Pass it as `-1` to force end the line with linebreak or as `-2` to force it without linebreak.
        """
        pass
    
    
    def add_token(self, token_type, token_value):
        """
        Adds a token to the context.
        
        Parameters
        ----------
        token_type : `int`
            The token's identifier.
        token_value : `None`, `str`
            The token's value.
        """
        token = Token(token_type, token_value)
        self.tokens.append(token)
    
    
    def add_tokens(self, tokens):
        """
        Adds tokens to the context.
        
        Parameters
        ----------
        tokens : `list` of ``Token``
            The tokens to add.
        """
        self.tokens.extend(tokens)
    
    
    def get_last_related_token(self):
        """
        Gets the last token of the highlight context.
        
        Returns
        -------
        token : ``Token``, `None`
            The token if there is any added.
        """
        tokens = self.tokens
        
        for token in reversed(tokens):
            token_type = token.type
            
            if token_type == TOKEN_TYPE_LINEBREAK_ESCAPED:
                continue
            
            if token_type == TOKEN_TYPE_SPACE:
                continue
            
            if token_type == TOKEN_TYPE_COMMENT:
                continue
            
            break
        else:
            token = None
        
        return token
    
    
    def match(self):
        """
        Matches the content of the context.
        """
        pass
    
    
    def generate_highlighted(self, formatter):
        """
        Generates highlighted content.
        
        This method is an iterable generator.
        
        Parameters
        ----------
        formatter : ``Formatter``
            Formatter to use for formatting content.
        
        Yields
        ------
        content : `str`
            The generated content.
        """
        return
        yield
    
    
    def _end_of_line(self):
        """
        Adds a linebreak token to the context.
        """
        tokens = self.tokens
        
        for token in reversed(tokens):
            token_type = token.type
            if token_type == TOKEN_TYPE_SPACE:
                continue
            
            if token_type == TOKEN_TYPE_COMMENT:
                continue
            
            if (token_type == TOKEN_TYPE_SPECIAL_OPERATOR) and (token.value == '\\'):
                token.type = TOKEN_TYPE_LINEBREAK_ESCAPED
            
            break


class HighlightParserContext(HighlightParserContextBase):
    """
    Represents a context of highlighting any content.
    
    Attributes
    ----------
    done : `bool`
        Whether processing is done.
    tokens : `list` of ``Token``
        The generated tokens.
    line_character_index : `int`
        The index of the character of the processed line.
    line_index : `int`
        The index of the line which is processed at the moment.
    lines : `list` of `str`
        The lines to highlight.
    """
    __slots__ = ('line_character_index', 'line_index', 'lines')
    
    def __new__(cls, lines):
        """
        Creates a new ``HighlightParserContext``.
        
        Parameters
        ----------
        lines : `list` of `str`
            The lines what the highlight context should match.
        """
        lines = [line for line in lines if line]
        
        if len(lines) == 0:
            done = True
        else:
            done = False
        
        self = object.__new__(cls)
        
        self.lines = lines
        self.line_index = 0
        self.line_character_index = 0
        self.done = done
        self.tokens = []
        
        return self
    
    
    @copy_docs(HighlightParserContextBase.get_line_index)
    def get_line_index(self):
        return self.line_index
    
    
    @copy_docs(HighlightParserContextBase.get_line)
    def get_line(self):
        lines = self.lines
        line_index = self.line_index
        if len(lines) <= line_index:
            line = ''
        else:
            line = lines[line_index]
        
        return line
    
    
    @copy_docs(HighlightParserContextBase.get_line_character_index)
    def get_line_character_index(self):
        return self.line_character_index
    
    
    @copy_docs(HighlightParserContextBase.set_line_character_index)
    def set_line_character_index(self, line_character_index):
        lines = self.lines
        line_index = self.line_index
        
        if line_index >= len(lines):
            line = ''
        else:
            line = lines[line_index]
        
        if (line_character_index < 0) or (len(line) <= line_character_index):
            line_index += 1
            
            self.line_index = line_index
            if len(lines) <= line_index:
                self.done = True
            
            if line_character_index != -2:
                self._end_of_line()
            
            line_character_index = 0
        
        self.line_character_index = line_character_index
    
    
    @copy_docs(HighlightParserContextBase.match)
    def match(self):
        while not self.done:
            for parser in PYTHON_PARSERS:
                if parser(self):
                    break
        
        tokens = self.tokens
        # Optimize tokens with merging sames into each other if applicable
        same_count = 0
        last_type = TOKEN_TYPE_ALL
        token_index = len(tokens) - 1
        
        while True:
            if token_index < 0:
                if same_count > 1:
                    # Merge tokens
                    _merge_tokens(tokens, 0, same_count)
                break
            
            token = tokens[token_index]
            token_type = token.type
            
            if (token_type not in MERGE_TOKEN_TYPES):
                last_type = token_type
                same_count = 0
                token_index -= 1
                continue
            
            if (last_type == token_type):
                token_index -= 1
                same_count += 1
                continue
            
            if same_count > 1:
                # Merge tokens
                _merge_tokens(tokens, token_index + 1, token_index + same_count + 1)
            
            same_count = 1
            last_type = token_type
            token_index -= 1
            continue
    
    
    @copy_docs(HighlightParserContextBase.generate_highlighted)
    def generate_highlighted(self, formatter):
        for token in self.tokens:
            yield from formatter.generate_highlighted(token)



@export
class FormatStringParserContext(HighlightParserContextBase):
    """
    Highlighter used to highlight format strings.
    
    Attributes
    ----------
    done : `bool`
        Whether processing is done.
    tokens : `list` of ``Token``
        The generated tokens.
    brace_level : `int`
        The internal brace level to un-match before entering a string.
    in_code : `bool`
        Whether we are parsing format code.
    line : `str`
        The internal content of the format string.
    line_character_index : `int`
        The index of the character of the processed line.
    """
    __slots__ = ('brace_level', 'in_code', 'line', 'line_character_index', )
    
    def __new__(cls, line):
        """
        Creates a new ``FormatStringParserContext``.
        
        Parameters
        ----------
        line : `str`
            A format string's internal content to highlight.
        """
        if len(line) == 0:
            done = True
        else:
            done = False
        
        self = object.__new__(cls)
        
        self.line = line
        self.done = done
        self.tokens = []
        self.line_character_index = 0
        self.brace_level = 0
        self.in_code = False
        
        return self
    
    
    @copy_docs(HighlightParserContextBase.get_line)
    def get_line(self):
        return self.line
    
    
    @copy_docs(HighlightParserContextBase.get_line_character_index)
    def get_line_character_index(self):
        return self.line_character_index
    
    
    @copy_docs(HighlightParserContextBase.set_line_character_index)
    def set_line_character_index(self, line_character_index):
        line = self.line
        
        if (line_character_index < 0) or (len(line) <= line_character_index):
            line_character_index = 0
            self.done = True
        
        self.line_character_index = line_character_index
    
    
    @copy_docs(HighlightParserContextBase.add_token)
    def add_token(self, token_type, token_value):
        if token_type == TOKEN_TYPE_LINEBREAK:
            self._end_of_line()
        
        # Check braces and such ~ Nya!
        elif token_type == TOKEN_TYPE_STRING_UNICODE_FORMAT:
            if self.in_code:
                token_type = TOKEN_TYPE_STRING_UNICODE_FORMAT_CODE
        
        elif token_type == TOKEN_TYPE_SPECIAL_PUNCTUATION and (token_value is not None):
            brace_level = self.brace_level
            if token_value == '{':
                if brace_level == self.in_code:
                    token_type = TOKEN_TYPE_STRING_UNICODE_FORMAT_MARK
                
                brace_level += 1
                self.brace_level = brace_level
            
            elif token_value == '}':
                if brace_level == 0:
                    # Random `}`- may be added as well, so if we are outside of f string internal, leave it alone.
                    token_type = TOKEN_TYPE_STRING_UNICODE_FORMAT
                else:
                    in_code = self.in_code
                    if brace_level <= in_code + 1:
                        token_type = TOKEN_TYPE_STRING_UNICODE_FORMAT_MARK
                        if (brace_level == 1) and in_code:
                            self.in_code = False
                    
                    brace_level -= 1
                    self.brace_level = brace_level
            
            elif token_value == ':':
                if (self.brace_level == 1) and (not self.in_code):
                    self.in_code = True
                    token_type = TOKEN_TYPE_STRING_UNICODE_FORMAT_MARK
        
        HighlightParserContextBase.add_token(self, token_type, token_value)
    
    
    @copy_docs(HighlightParserContextBase.match)
    def match(self):
        while True:
            if self.done:
                break
            
            if (self.brace_level == self.in_code):
                _try_match_till_format_string_expression(self)
            else:
                while True:
                    for parser in PYTHON_PARSERS_FORMAT_STRING:
                        if parser(self):
                            break
                    
                    # Need goto!
                    if self.done:
                        break
                    
                    # Need goto!
                    if self.brace_level == self.in_code:
                        break



__all__ = ('ResultGatheringFuture',)

import reprlib

from ...utils.trace import format_callback

from ..exceptions import InvalidStateError

from .future import FUTURE_STATE_FINISHED, FUTURE_STATE_PENDING, Future


class ResultGatheringFuture(Future):
    """
    A Future subclass, which yields after it's result was set a set amount of times with ``.set_result``, or with
    ``.set_result_if_pending``, or till an exception is set to with ``.set_exception``, or with
    ``.set_exception_if_pending``.
    
    Attributes
    ----------
    _blocking : `bool`
        Whether the future is already being awaited, so it blocks the respective coroutine.
    _callbacks : `list` of `callable`
        The callbacks of the future, which are queued up on the respective event loop to be called, when the future is
        finished. These callback should accept `1` parameter, the future itself.
        
        Note, if the future is already done, then the newly added callbacks are queued up instantly on the respective
        event loop to be called.
    
    _exception : `None`, `BaseException`
        The exception set to the future as it's result. Defaults to `None`.
    _loop : ``EventThread``
        The loop to what the created future is bound.
    _result : `list` of `Any`
        The results of the future.
    _state : `str`
        The state of the future.
        
        Can be set as one of the following:
        
        +---------------------------+-----------+
        | Respective name           | Value     |
        +===========================+===========+
        | FUTURE_STATE_PENDING      | `0`       |
        +---------------------------+-----------+
        | FUTURE_STATE_CANCELLED    | `1`       |
        +---------------------------+-----------+
        | FUTURE_STATE_FINISHED     | `2`       |
        +---------------------------+-----------+
        | FUTURE_STATE_RETRIEVED    | `3`       |
        +---------------------------+-----------+
        
        Note, that states are checked by memory address and not by equality. Also ``FUTURE_STATE_RETRIEVED`` is used
        only if `__debug__` is set as `True`.
    _count : `int`
        The amount, how much times the future's result need to be set, because it will yield.
    """
    __slots__ = ('_count',)
    
    def __new__(cls, loop, count):
        """
        Creates a new ``ResultGatheringFuture`` object bound to the given `loop`, which will be marked as done, only if `count`
        results are set to it with ``.set_result``, or with ``.set_result_if_pending``.
        
        Parameters
        ----------
        loop : ``EventThread``
            The loop to what the created future will be bound to.
        count : `int`
            The amount of times, the future's result need to be set, because becoming done.
        """
        self = object.__new__(cls)
        self._loop = loop
        self._count = count
        self._state = FUTURE_STATE_PENDING
        
        self._result = []
        self._exception = None
        
        self._callbacks = []
        self._blocking = False
        
        return self
    
    
    def __repr__(self):
        """Returns the gatherer's representation."""
        repr_parts = ['<', self.__class__.__name__, ' ']
        
        state = self._state
        repr_parts.append(state)
        
        if state >= FUTURE_STATE_FINISHED:
            exception = self._exception
            if exception is None:
                results = self._result
                for index, result in enumerate(results):
                    repr_parts.append(f', result[')
                    repr_parts.append(repr(index))
                    repr_parts.append(']=')
                    repr_parts.append(reprlib.repr(result))
                
                repr_parts.append(', needed=')
                repr_parts.append(str(self._count - len(results)))
            else:
                repr_parts.append(', exception=')
                repr_parts.append(repr(exception))
        
        callbacks = self._callbacks
        limit = len(callbacks)
        if limit:
            repr_parts.append(', callbacks=[')
            index = 0
            while True:
                callback = callbacks[index]
                repr_parts.append(format_callback(callback))
                index += 1
                if index == limit:
                    break
                
                repr_parts.append(', ')
                continue
            
            repr_parts.append(']')
        
        repr_parts.append('>')
        
        return ''.join(repr_parts)
    
    
    def set_result(self, result):
        """
        Sets the future result, and if it waits for no more results, marks it as done as well.
        
        Parameters
        ----------
        result : `Any`
            The object to set as result.
        
        Raises
        ------
        InvalidStateError
            If the future is already done.
        """
        if self._state != FUTURE_STATE_PENDING:
            raise InvalidStateError(self, 'set_result')
        
        results = self._result
        results.append(result)
        if self._count != len(results):
            return
        
        self._state = FUTURE_STATE_FINISHED
        self._loop._schedule_callbacks(self)


    def set_result_if_pending(self, result):
        """
        Sets the future result, and if it waits for no more results, marks it as done as well. Not like
        ``.set_result``, this method will not raise ``InvalidStateError`` if the future is already done.
        
        Parameters
        ----------
        result : `Any`
            The object to set as result.
        
        Returns
        ------
        set_result : `int` (`0`, `1`, `2`)
            If the future is already done, returns `0`. If the future's result was successfully set, returns `1`,
            meanwhile if the future was marked as done as well, returns `2`.
        """
        if self._state != FUTURE_STATE_PENDING:
            return 0
        
        results = self._result
        results.append(result)
        if self._count != len(results):
            return 1
            
        self._state = FUTURE_STATE_FINISHED
        self._loop._schedule_callbacks(self)
        return 2
    
    
    def clear(self):
        """
        Clears the future making it reusable.
        """
        self._state = FUTURE_STATE_PENDING
        self._exception = None
        self._result.clear()
        self.cancel_handles()
        self._blocking = False




__all__ = ()

class RequestInfo:
    """
    Base information representing a request.
    
    Attributes
    ----------
    headers : ``IgnoreCaseMultiValueDictionary``
        The respective request's headers.
    method : `str`
        The respective request's method.
    real_url : ``URL``
        The url given to request.
    url : ``URL``
        The requested url without fragments. Can be same as ``.real_url``.
    """
    __slots__ = ('headers', 'method', 'real_url', 'url',)
    
    def __init__(self, request):
        """
        Creates a new ``RequestInfo`` representing the given request.
        
        Parameters
        ----------
        request : ``ClientRequest``
            The represented request.
        """
        self.url = request.url
        self.method = request.method
        self.headers = request.headers
        self.real_url = request.original_url
    
    def __repr__(self):
        """Returns the request info's representation."""
        return f'<{self.__class__.__name__} url={self.url!r}>'



__all__ = ('MethodLike',)

from types import MethodType

from .docs import has_docs


@has_docs
class SubCheckType(type):
    """
    Metaclass, which can be used for subclass checks. It's type instances should implement a `.__subclasses__`
    class attribute, which contains all of it's "subclasses".
    """
    @has_docs
    def __instancecheck__(cls, instance):
        """Returns whether the given instance's type is a subclass of the respective type."""
        return (type(instance) in cls.__subclasses__)
    
    @has_docs
    def __subclasscheck__(cls, klass):
        """Returns whether the given type is a subclass of the respective type."""
        return (klass in cls.__subclasses__)


@has_docs
class MethodLike(metaclass=SubCheckType):
    """
    Base class for methods.
    
    Class Attributes
    ----------------
    __subclasses__ : `set` of `type`
        Registered method types.
    __reserved_parameter_count__ : `int` = `1`
        The amount of reserved parameters by a method subclass.
    """
    __subclasses__ = {MethodType}
    __slots__ = ()
    def __init_subclass__(cls):
        cls.__subclasses__.add(cls)
    
    __reserved_parameter_count__ = 1
    
    @classmethod
    @has_docs
    def get_reserved_parameter_count(cls, instance):
        """
        Returns the given `instance`'s reserved argcount.
        
        Parameters
        ----------
        instance : `method-like`
            A method like object.
        
        Returns
        -------
        reserved_parameter : `int`
            Reserved argcount of the given method like.
        
        Raises
        ------
        TypeError
            If `instance` is not given as a `method-like`.
        """
        instance_type = instance.__class__
        reserved_parameter = getattr(instance_type, '__reserved_parameter_count__', -1)
        if reserved_parameter != -1:
            return reserved_parameter
        
        if instance_type in cls.__subclasses__:
            return cls.__reserved_parameter_count__
        
        raise TypeError(
            f'Expected a method like, got {instance_type.__name__}; {instance!r}.'
        )



__all__ = ()

# import sub-modules

from .compilation import *
from .editor_base import *
from .editor_simple import *
from .prefix_trimming import *

# only import advanced editor when it supposed to work

try:
    import termios
    import tty
except ImportError:
    editor_advanced = None
else:
    from .editor_advanced import *

# collect editors

from .editor_base import EditorBase

from .editor_simple import EditorSimple

if (editor_advanced is None):
    EditorAdvanced = None
else:
    from .editor_advanced import EditorAdvanced

# define `can_use_advanced_editor`

from os import get_terminal_size
from sys import platform as PLATFORM

def can_use_advanced_editor():
    if (EditorAdvanced is None):
        return False
    
    if PLATFORM != 'linux':
        return False
    
    try:
        get_terminal_size()
    except OSError:
        return False
    
    return True



__all__ = ()

from html import escape as html_escape

from .token_types import TOKEN_TYPE_LINEBREAK, TOKEN_TYPE_LINEBREAK_ESCAPED


class FormatterDetail:
    """
    Represents a formatter node's detail.
    
    Attributes
    ----------
    details : `Any`
        Additional details to pass to the ``.formatter_generator_function``.
    formatter_generator_function : ``GeneratorFunction``
        Generator function which
    """
    __slots__ = ('details', 'formatter_generator_function')
    
    def __new__(cls, formatter_generator_function, details):
        """
        Creates a new formatter detail.
        
        Parameters
        ----------
        formatter_generator_function : ``GeneratorFunction``
            Generator function which
        details : `Any`
            Additional details to pass to the ``.formatter_generator_function``.
        """
        self = object.__new__(cls)
        self.details = details
        self.formatter_generator_function = formatter_generator_function
        return self
    
    
    def __repr__(self):
        """Returns the formatter detail's representation."""
        return f'{self.__class__.__name__}({self.formatter_generator_function!r}, {self.details!r})'
    
    
    def __call__(self, token):
        """
        Generates highlighted content from the given token.
        
        This method is an iterable generator.
        
        Parameters
        ----------
        token : ``Token``
            The token to generate it's highlighted version.
        
        Yields
        ------
        content : `str`
        """
        yield from self.formatter_generator_function(self.details, token)
    
    
    def __eq__(self, other):
        """Returns whether the two formatter details are equal."""
        if type(self) is not type(other):
            return NotImplemented
        
        if self.details != other.details:
            return False
        
        if self.formatter_generator_function is not other.formatter_generator_function:
            return False
        
        return True


def formatter_html(html_class, token):
    """
    Html highlight text formatter.
    
    Parameters
    ----------
    html_class : `None`, `str`
        The html class to format the string as.
    token : ``Token``
        The token to generate it's highlighted version.
        
    Yields
    ------
    content : `str`
    """
    token_type = token.type
    
    token_value = token.value
    if (token_value is not None):
        if (html_class is not None):
            yield '<span class="'
            yield html_class
            yield '">'
        
        yield html_escape(token_value)
        
        if (html_class is not None):
            yield '</span>'
    
    if (token_type == TOKEN_TYPE_LINEBREAK):
        yield '<br>'


def formatter_ansi_code(format_and_reset_code, token):
    """
    Ansi highlight text formatter.
    
    Parameters
    ----------
    format_and_reset_code : `None`, `tuple` (`str`, `str`)
        Ansi format code and reset code.
    token : ``Token``
        The token to generate it's highlighted version.
        
    Yields
    ------
    content : `str`
    """
    token_value = token.value
    if (token_value is None):
        return
    
    if format_and_reset_code is None:
        yield token_value
        return
    
    starter, ender = format_and_reset_code
    yield starter
    yield token_value
    yield ender
    return



__all__ = ('iter_highlight_code_lines',)

from .parser_context import HighlightParserContext


def iter_highlight_code_lines(lines, formatter_context):
    """
    Matches the given python code lines and iterates it's formatted representation.
    
    This function is an iterable generator.
    
    Parameters
    ----------
    lines : `list` of `str`
        Lines to format.
    formatter_context : ``HighlightFormatterContext``
        Context to use for highlighting.
    
    Yields
    ------
    content : `str`
    """
    context = HighlightParserContext(lines)
    context.match()
    yield from context.generate_highlighted(formatter_context)



__all__ = ()

def _dict_update_iterable_iterator(iterable):
    """
    Iterates over the elements of the given `iterable`. With each loop yields 2 elements, a key-value pair.
    
    This method is an iterable generator.
    
    Parameters
    ----------
    iterable : `iterable`
        An iterable to add it's elements to a list.
    
    Yields
    ------
    item : `tuple` (`object`, `object`)
    
    Raises
    ------
    TypeError
        An element's type is incorrect.
    """
    index = - 1
    for item in iterable:
        index += 1
        
        try:
            iterator = iter(item)
        except TypeError:
            raise TypeError(
                f'Cannot convert dictionary update sequence element #{index} to a sequence; '
                f'got item={item!r}; iterable={iterable!r}.'
            ) from None
        
        try:
            key = next(iterator)
        except StopIteration:
            raise ValueError(
                f'Dictionary update sequence element #{index} has length `0`; 2 is required; '
                f'got item={item!r}; iterable={iterable!r}.'
            ) from None
        
        try:
            value = next(iterator)
        except StopIteration:
            raise ValueError(
                f'Dictionary update sequence element #{index} has length `1`; 2 is required; '
                f'got item={item!r}; iterable={iterable!r}.'
            ) from None
        
        try:
            next(iterator)
        except StopIteration:
            pass
        else:
            length = 3
            for _ in iterator:
                length += 1
                if length > 9000:
                    break
            
            if length > 9000:
                length = 'OVER 9000!'
            else:
                length = repr(length)
            
            raise ValueError(
                f'Dictionary update sequence element #{index} has length {length}; 2 is required; '
                f'got item={item!r}; iterable={iterable!r}.'
            )
        
        yield key, value



__all__ = ('MimeType',)

from ..utils import MultiValueDictionary


class MimeType:
    # Parses a MIME type into its components
    __slots__ = ('parameters', 'sub_type', 'suffix', 'type', )
    
    def __init__(self, mime_type):
        
        if (mime_type is None) or (not mime_type):
            self.type = ''
            self.sub_type = ''
            self.suffix = ''
            self.parameters = {}
            return
        
        parts = mime_type.split(';')
        parameters = MultiValueDictionary()
        for item in parts[1:]:
            if not item:
                continue
            
            if '=' in item:
                key, value = item.split('=', 1)
            else:
                key = item
                value = ''
            
            parameters[key.strip().lower()] = value.strip(' "')
        
        
        full_type = parts[0].strip().lower()
        if full_type == '*':
            full_type = '*/*'
        
        if '/' in full_type:
            type_, sub_type = full_type.split('/', 1)
        else:
            type_ = full_type
            sub_type = ''

        if '+' in sub_type:
            sub_type, suffix = sub_type.split('+', 1)
        else:
            suffix = ''
            
        self.type = type_
        self.sub_type = sub_type
        self.suffix = suffix
        self.parameters = parameters
    
    
    def __repr__(self):
        return (
            f'<{self.__class__.__name__} type={self.type!r} sub_type={self.sub_type!r} suffix={self.suffix!r} '
            f'parameters={self.parameters!r}>'
        )



__all__ = ('HighlightFormatterContext',)

from .ansi import create_ansi_format_code
from .formatter_detail import FormatterDetail, formatter_ansi_code, formatter_html
from .formatter_node import FormatterNode

from .token import Token
from .token_types import TOKEN_STRUCTURE

ANSI_RESET_CODE = create_ansi_format_code()


class HighlightFormatterContext:
    """
    Formatter context which holds details for formatting.
    
    Attributes
    ----------
    formatter_nodes : `dict` of (`int, ``FormatterNode``) items
        The registered formatter nodes.
    """
    __slots__ = ('formatter_nodes',)
    
    def __new__(cls):
        """
        Creates a new formatter context.
        """
        self = object.__new__(cls)
        self.formatter_nodes = {}
        
        self._build_type_token_nodes(TOKEN_STRUCTURE)
        
        return self
    
    
    def _build_type_token_nodes(self, dictionary):
        """
        Builds a token node type structure from the given dictionary.
        
        Parameters
        ----------
        dictionary : `dict` of (`int`, `None` or repeat) items
            A dictionary, which describes the node structure.
        """
        for key, value in dictionary.items():
            node = FormatterNode(self, key)
            if (value is not None):
                self._build_type_token_child(node, value)
    
    
    def _build_type_token_child(self, parent, dictionary):
        """
        Builds token node type structure extending the given parent.
        
        Parameters
        ----------
        parent : ``FormatterNode``
            The parent node.
        dictionary : `dict` of (`int`, `None` or repeat) items
            A dictionary, which describes the node structure.
        """
        for key, value in dictionary.items():
            node = FormatterNode(self, key)
            parent.add_node(node)
            if (value is not None):
                self._build_type_token_child(node, value)
    

    def __repr__(self):
        """Returns the formatter's representation."""
        return f'<{self.__class__.__name__} nodes: {len(self.formatter_nodes)}>'
    
    
    def highlight_as(self, content, token_type):
        """
        Highlights the given text as it would be a token of the given type.
        
        Parameters
        -----------
        content : `str`
            The content to highlight.
        token_type : `int`
            The token type to use.
        """
        detail = self.formatter_nodes[token_type].detail
        if (detail is None):
            return content
        
        else:
            return ''.join([*detail(Token(token_type, content))])
    
    
    def generate_highlighted(self, token):
        """
        Generates highlighted content from the given token.
        
        This method is an iterable generator.
        
        Parameters
        ----------
        token : ``Token``
            The token to generate it's highlighted version.
        
        Yields
        ------
        content : `str`
        """
        detail = self.formatter_nodes[token.type].detail
        if (detail is None):
            value = token.value
            if (value is not None):
                yield value
        
        else:
            yield from detail(token)
    
    
    def set_highlight_html_all(self):
        """
        Sets html formatting for all nodes of the context.
        """
        detail = FormatterDetail(formatter_html, None)
        for node in self.formatter_nodes.values():
            node.set_detail(detail, direct=False)
            
    
    def set_highlight_html_class(self, token_type_identifier, html_class):
        """
        Sets html class formatting for the given node.
        
        Parameters
        ----------
        token_type_identifier : `int`
            The node's identifier.
        html_class : `None`, `str`
            The html class to set.
        
        Raises
        ------
        TypeError
            - If `token_type_identifier` was not given as `int`.
            - If `html_class` was not given neither as `None` nor as `str`.
        ValueError
            If `token_type_identifier` was not given as any of the predefined values. Check ``TOKEN_TYPES`` for more
                details.
        """
        if (html_class is not None) and (not isinstance(html_class, str)):
            raise TypeError(
                f'`html_class` can be `None`, `str`, got {html_class.__class__.__name__}; {html_class!r}.'
            )
        
        self.set_highlight_detail(
            token_type_identifier,
            FormatterDetail(formatter_html, html_class)
        )
    
    
    def set_highlight_ansi_code(self, token_type_identifier, ansi_code):
        """
        Sets ansi formatting for the given node.
        
        Parameters
        ----------
        token_type_identifier : `int`
            The node's identifier.
        ansi_code : `None`, `str`
            The format code to set.
        
        Raises
        ------
        TypeError
            - If `token_type_identifier` was not given as `int`.
            - If `ansi_code` was not given neither as `None` nor as `str`.
        ValueError
            If `token_type_identifier` was not given as any of the predefined values. Check ``TOKEN_TYPES`` for more
                details.
        """
        if (ansi_code is not None) and (not isinstance(ansi_code, str)):
            raise TypeError(
                f'`ansi_code` can be `None`, `str`, got {ansi_code.__class__.__name__}; {ansi_code!r}.'
            )
        
        self.set_highlight_detail(
            token_type_identifier,
            FormatterDetail(formatter_ansi_code, (ansi_code, ANSI_RESET_CODE))
        )
    
    
    def set_highlight_detail(self, token_type_identifier, detail):
        """
        Sets detail for the given node.
        
        Parameters
        ----------
        token_type_identifier : `int`
            The node's identifier.
        detail : `None`, ``FormatterDetail``
            The detail to set.
        
        Raises
        ------
        TypeError
            - If `token_type_identifier` was not given as `int`.
        ValueError
            If `token_type_identifier` was not given as any of the predefined values. Check ``TOKEN_TYPES`` for more
                details.
        """
        if not isinstance(token_type_identifier, int):
            raise TypeError(
                f'`token_type_identifier` can be `int`, got '
                f'{token_type_identifier.__class__.__name__}; {token_type_identifier!r}.'
            )
        
        try:
            node = self.formatter_nodes[token_type_identifier]
        except KeyError:
            raise ValueError(
                f'`token_type_identifier` was not given as any of the predefined values, got '
                f'{token_type_identifier!r}.'
            ) from None
        
        node.set_detail(detail)



__all__ = ()

EXTRA_INFO_NAME_SSL_CONTEXT = 'ssl_context'
EXTRA_INFO_NAME_SSL_OBJECT = 'ssl_object'
EXTRA_INFO_NAME_SOCKET = 'socket'
EXTRA_INFO_NAME_PEER_CERTIFICATION = 'peer_certification'
EXTRA_INFO_NAME_CIPHER = 'cipher'
EXTRA_INFO_NAME_COMPRESSION = 'compression'
EXTRA_INFO_NAME_SOCKET_NAME = 'socket_name'
EXTRA_INFO_NAME_PEER_NAME = 'peer_name'
EXTRA_INFO_NAME_PIPE = 'pipe'

ALTERNATIVE_EXTRA_INFO_NAMES = {
    EXTRA_INFO_NAME_SSL_CONTEXT: 'ssl_context',
    EXTRA_INFO_NAME_SOCKET: 'sock',
    EXTRA_INFO_NAME_PEER_CERTIFICATION: 'peercert',
    EXTRA_INFO_NAME_SOCKET_NAME: 'sockname',
    EXTRA_INFO_NAME_PEER_NAME: 'peername',
}

def get_extra_info(extra, name, default):
    """
    Gets optional transport information.
    
    Parameters
    ----------
    extra : `None`, `dict` of (`str`, `Any`) items
        Optional transform information.
    name : `str`
        The extra information's name to get.
    default : `Any`
        Default value to return if `name` could not be matched. Defaults to `None`.
    
    Returns
    -------
    info : `default`, `Any`
    """
    if extra is None:
        info = None
    else:
        try:
            info = extra[name]
        except KeyError:
            try:
                name = ALTERNATIVE_EXTRA_INFO_NAMES[name]
            except KeyError:
                info = default
            else:
                info = extra.get(name, default)
    
    return info


def set_extra_info(extra, name, value):
    """
    Sets optional transport information.
    
    Parameters
    ----------
    extra : `None`, `dict` of (`str`, `Any`) items
        Optional transform information.
    name : `str`
        The extra info's name.
    value : `Any`
        The value to set.
    
    Returns
    -------
    extra : `dict` of (`str`, `Any`) items
        The new transform information.
    """
    if extra is None:
        extra = {}
    
    extra[name] = value
    
    return extra


def has_extra_info(extra, name):
    """
    Checks optional transport information.
    
    Parameters
    ----------
    extra : `None`, `dict` of (`str`, `Any`) items
        Optional transform information.
    name : `str`
        The extra info's name.
    
    Returns
    -------
    has_extra_info : `bool`
        Whether the transport has the extra information.
    """
    if extra is None:
        has_extra_info = False
    else:
        if name in extra:
            has_extra_info = True
        else:
            try:
                name = ALTERNATIVE_EXTRA_INFO_NAMES[name]
            except KeyError:
                has_extra_info = False
            else:
                has_extra_info = (name in extra)
    
    return has_extra_info


def get_has_extra_info(extra, name):
    """
    Gets optional transport information and whether it is present, or nah.
    
    Parameters
    ----------
    extra : `None`, `dict` of (`str`, `Any`) items
        Optional transform information.
    name : `str`
        The extra information's name to get.
    
    Returns
    -------
    info : `default`, `Any`
        Optional transport information.
    present : `bool`
        Whether the field is present.
    """
    if extra is None:
        info = None
        present = False
    else:
        try:
            info = extra[name]
        except KeyError:
            try:
                name = ALTERNATIVE_EXTRA_INFO_NAMES[name]
            except KeyError:
                info = None
                present = False
            else:
                try:
                    info = extra[name]
                except KeyError:
                    info = None
                    present = False
                else:
                    present = True
        
        else:
            present = True
        
    return info, present



import sys
from os import getcwd as get_current_work_directory
from os.path import (
    dirname as get_directory_name, expanduser as get_user_home_directory, join as join_paths,
    normpath as normalize_path, realpath as get_real_path
)


try:
    from . import __package__ as PACKAGE_NAME
except ImportError:
    PACKAGE_NAME = sys.path[0]
    
    sys.path.append(
        normalize_path(
            join_paths(
                get_directory_name(
                    get_real_path(
                        join_paths(
                            get_current_work_directory(),
                            get_user_home_directory(__file__),
                        )
                    )
                ),
                '..',
            )
        )
    )


PACKAGE = __import__(f'{PACKAGE_NAME}.tools.asynchronous_interactive_console')

collect_package_local_variables = PACKAGE.tools.collect_package_local_variables
run_asynchronous_interactive_console = \
    PACKAGE.tools.asynchronous_interactive_console.run_asynchronous_interactive_console



def __main__():
    """
    Runs an asynchronous interactive console.
    
    > You need python3.8 or higher to use `await`!
    """
    run_asynchronous_interactive_console(collect_package_local_variables(PACKAGE))


if __name__ == '__main__':
    __main__()



__all__ = ('quote', 'unquote',)

from string import ascii_letters, ascii_lowercase, digits


BINARY_ASCII_LOWERCASE = frozenset(ascii_lowercase.encode('ascii'))
BINARY_PERCENTAGE_ALLOWED = frozenset((f'%{index:02X}'.encode('ascii') for index in range(256)))
GEN_DELIMITERS = frozenset(':/?#[]@')
SUB_DELIMITERS_WITHOUT_QUOTES = frozenset('!$\'()*,;')
QUERY_STRING_NOT_SAFE = frozenset('+&=')
SUB_DELIMITERS = frozenset((*SUB_DELIMITERS_WITHOUT_QUOTES, *QUERY_STRING_NOT_SAFE))
RESERVED = frozenset((*GEN_DELIMITERS, *SUB_DELIMITERS))
UNRESERVED = frozenset((*ascii_letters, *digits, *'-._~'))
ALLOWED = frozenset((*UNRESERVED, *SUB_DELIMITERS_WITHOUT_QUOTES))

QUOTE_SAFE_MAP = {}
UNQUOTE_UNSAFE_MAP = {}

def iter_nullable_string(string):
    """
    Iterates overt he given nullable string.
    
    Parameters
    ----------
    string : `None`, `str`
        The string to iterate trough
    
    Yields
    ------
    element : `str`
    """
    if (string is not None):
        yield from string


def quote(value, safe=None, protected=None, query_string=False):
    """
    Http quotes the given `value`.
    
    Parameters
    ----------
    value : `None`, `str`
        The value to quote.
    safe : `None`, `str` = `None`, Optional
        Additional not percentage encoding safe characters, which should not be contained by potentially percent
        encoded characters.
    protected : `None`, `str` = `None`, Optional
        Additional characters, which should not be percentage encoded.
    query_string : `bool` = `False`, Optional
        Whether the `value` is a query string value.
    
    Returns
    -------
    unquoted : `None`, `str`
        The unquoted value. Returns `None` of `value` was given as `None` as well.
    
    Raises
    ------
    TypeError
        If `value` was not given neither as `None`, nor `str`.
    """
    if value is None:
        return None
    
    if not isinstance(value, str):
        raise TypeError(
            f'`value` can be `None`, `str`, got {value.__class__.__name__}; {value!r}.'
        )
    
    if not value:
        return ''
    
    value = value.encode('utf8')
    result = bytearray()
    percentage = None
    
    key = (safe, protected, query_string)
    try:
        safe, binary_safe = QUOTE_SAFE_MAP[key]
    except KeyError:
        safe = frozenset((
            *iter_nullable_string(safe),
            *ALLOWED,
            *iter_nullable_string(protected),
            *iter_nullable_string(None if query_string else QUERY_STRING_NOT_SAFE),
        ))
        
        binary_safe = frozenset(ord(char) for char in safe)
        
        QUOTE_SAFE_MAP[key] = (safe, binary_safe)
    
    
    for char in value:
        if (percentage is not None):
            if char in BINARY_ASCII_LOWERCASE:
                char -= 32
            
            percentage.append(char)
            if len(percentage) == 3:
                percentage = bytes(percentage)
                try:
                    unquoted = chr(int(percentage[1:].decode('ascii'), base=16))
                except ValueError as err:
                    raise ValueError(
                        f'Unallowed percentage: {percentage!r}; value={value!r}.'
                    ) from err
                
                if (protected is not None) and (unquoted in protected):
                    result.extend(percentage)
                elif unquoted in safe:
                    result.append(ord(unquoted))
                else:
                    result.extend(percentage)
                
                percentage = None
            
            continue
            
        if char == b'%'[0]:
            percentage = bytearray()
            percentage.append(char)
            continue
        
        if query_string:
            if char == b' '[0]:
                result.append(b'+'[0])
                continue
        
        if char in binary_safe:
            result.append(char)
            continue
        
        result.extend((f'%{char:02X}').encode('ascii'))
    
    return result.decode('ascii')


def unquote(value, unsafe=None, query_string=False):
    """
    Http unquotes the given `value`.
    
    Parameters
    ----------
    value : `None`, `str`
        The value to quote.
    unsafe : `None`, `str` = `None`, Optional
        Additional not percentage encoding safe characters, which should not be contained by potentially percent
        encoded characters.
    query_string : `bool` = `False`, Optional
        Whether the `value` is a query string value.
    
    Returns
    -------
    unquoted : `None`, `str`
        The unquoted value. Returns `None` of `value` was given as `None` as well.
    
    Raises
    ------
    TypeError
        If `value` was not given neither as `None`, nor `str`.
    """
    if value is None:
        return None
    
    if not isinstance(value, str):
        raise TypeError(
            f'`value` can be `None`, `str`, got {value.__class__.__name__}; {value!r}.'
        )
    
    if not value:
        return ''
    
    percentage = None
    last_percentage = ''
    percentages = bytearray()
    result = []
    
    if (unsafe is not None):
        try:
            unsafe = UNQUOTE_UNSAFE_MAP[unsafe]
        except KeyError:
            unsafe = UNQUOTE_UNSAFE_MAP[unsafe] = frozenset(unsafe)
    
    for char in value:
        if (percentage is not None):
            percentage += char
            if len(percentage) == 3:
                percentages.append(int(percentage[1:], base=16))
                last_percentage = percentage
                percentage = None
            continue
        
        if percentages:
            try:
                unquoted = percentages.decode('utf8')
            except UnicodeDecodeError:
                pass
            else:
                if query_string and (unquoted in QUERY_STRING_NOT_SAFE):
                    result.append(quote(unquoted, query_string=True))
                elif (unsafe is not None) and (unquoted in unsafe):
                    result.append(quote(unquoted))
                else:
                    result.append(unquoted)
                percentages.clear()
        
        if char == '%':
            percentage = char
            continue
        
        if percentages:
            result.append(last_percentage)  # %F8ab
            last_percentage = ''
        
        if char == '+':
            if (unsafe is None) or (char not in unsafe):
                char = ' '
            
            result.append(char)
            continue
        
        if (unsafe is not None) and (char in unsafe):
            result.append('%')
            result.extend(ord(char).__format__('X'))
            continue
        
        result.append(char)
    
    if percentages:
        try:
            unquoted = percentages.decode('utf8')
        except UnicodeDecodeError:
            result.append(last_percentage)  # %F8
        else:
            if query_string and (unquoted in QUERY_STRING_NOT_SAFE):
                result.append(quote(unquoted, query_string=True))
            elif (unsafe is not None) and (unquoted in unsafe):
                result.append(quote(unquoted))
            else:
                result.append(unquoted)
    
    return ''.join(result)



from .websocket_client import *
from .websocket_common_protocol import *
from .websocket_server import *
from .websocket_server_protocol import *

__all__ = (
    *websocket_client.__all__,
    *websocket_common_protocol.__all__,
    *websocket_server.__all__,
    *websocket_server_protocol.__all__,
)

# Resolve import if http_client is not imported
from .. import http_client



from .client_request import *
from .client_response import *
from .connection import *
from .connection_key import *
from .connector import *
from .fingerprint import *
from .http_client import *
from .request_context_managers import *
from .request_info import *

__all__ = (
    *client_request.__all__,
    *client_response.__all__,
    *connection.__all__,
    *connection_key.__all__,
    *connector.__all__,
    *fingerprint.__all__,
    *http_client.__all__,
    *request_context_managers.__all__,
    *request_info.__all__,
)



__all__ = ('KeepType',)

from .docs import has_docs


@has_docs
class KeepType:
    """
    A decorator, what can be used to add features to an already existing class, by defining a new one, what will extend
    the old one's functionality.
    
    Note, that already existing attributes will not be overwritten and neither of the followingly named attributes
    either:
        - `__name__`
        - `__qualname__`
        - `__weakref__`
        - `__dict__`
        - `__slots__`
    
    Attributes
    ----------
    old_class : `type`
        The old class to extend.
    
    Class Attributes
    ----------------
    _ignored_attr_names : `set` of `str`
        Attribute names to ignore when extending.
    """
    __slots__ = ('old_class',)
    
    _ignored_attr_names = frozenset(('__name__', '__qualname__', '__weakref__', '__dict__', '__slots__', '__module__'))
    
    @has_docs
    def __new__(cls, old_class, *, new_class=None):
        """
        Creates a new ``KeepType`` with given `old_class` to extend. Can be used as a decorator if `new_class`
        parameter is not given.
        
        Parameters
        ----------
        old_class : `type`
            The old class to extend.
        new_class : `None`, `type` = `None`, Optional (Keyword only)
            The new class to extend the old class's functionality with.
        
        Returns
        -------
        obj : ``KeepType``, `type`
            If only `old_class` attribute is given, then returns itself enabling using it as a decorator, but if
            `new_class` is given as well, then returns the extended `old_class`.
        """
        self = object.__new__(cls)
        self.old_class = old_class
        
        if new_class is None:
            return self
        
        return self(new_class)
    
    @has_docs
    def __call__(self, new_class):
        """
        Calls the ``KeepType`` extending it's old ``.old_class`` with the new given `new_class`.
        
        Parameters
        ----------
        new_class : `type`
            The new class to extend the old class's functionality with.
        
        Returns
        -------
        old_class : `type`
            The extended old class.
        """
        old_class = self.old_class
        ignored_attr_names = self._ignored_attr_names
        for attribute_name in dir(new_class):
            if attribute_name in ignored_attr_names:
                continue
            
            attribute_value = getattr(new_class, attribute_name)
            if (attribute_name == '__doc__') and (attribute_value is None):
                continue
            
            if hasattr(object, attribute_name) and (attribute_value is getattr(object, attribute_name)):
                continue
            
            setattr(old_class, attribute_name, attribute_value)
        
        return old_class



__all__ = ('CancelledError', 'InvalidStateError',)

from ..utils import include


get_future_state_name = include('get_future_state_name')


class CancelledError(BaseException):
    """The Future or Task was cancelled."""


class InvalidStateError(Exception):
    """
    The operation is not allowed in this state.
    
    Attributes
    ----------
    future : ``Future``.
        The future, from what's method the exception was raised.
    func_name : `str`
        The future's function's name, from where the exception was raised.
    _message : `None`, `str`
        Internal cache for the ``.message`` property.
    """
    def __init__(self, future, func_name, message=None):
        """
        Creates a new ``InvalidStateError``.
        
        Parameters
        ----------
        future : ``Future``.
            The future, from what's method the exception was raised.
        func_name : `str`
            The future's function's name, from where the exception was raised.
        message : `str`
            The exception's message. If not defined, then the exception will generate it by itself.
        """
        self.future = future
        self.func_name = func_name
        self._message = message
    
    def __repr__(self):
        """Returns the exception's representation."""
        return f'{self.__class__.__name__}: {self.message}'
    
    __str__ = __repr__
    
    @property
    def message(self):
        """
        Returns the exception's message.
        
        Returns
        -------
        message : `str`
        
        Notes
        -----
        If the exception was created wia ``.__init__``, without giving defining the `message` parameter, then the
        exception's message is generated only when this property is retrieved for the first time.
        """
        message = self._message
        if message is None:
            future = self.future
            message = (
                f'`{future.__class__.__name__}.{self.func_name}` was called, when `.state` is {future._state} '
                f'({get_future_state_name(future._state)}) of {future!r}.'
            )
            self._message = message
        
        return message



__all__ = ('run_asynchronous_interactive_console',)

import sys, warnings
from functools import partial as partial_func
from math import floor, log
from types import FunctionType

from ...core import Future, get_default_trace_writer_highlighter, get_event_loop
from ...utils import HIGHLIGHT_TOKEN_TYPES, is_awaitable, render_exception_into
from ...utils.trace import (
    _render_syntax_error_representation_into, fixup_syntax_error_line_from_buffer, is_syntax_error
)

from .console_helpers import create_banner, create_exit_message, get_or_create_event_loop
from .editors import EditorAdvanced, EditorBase, EditorSimple, can_use_advanced_editor
from .history import History


def _ignore_console_frames(file_name, name, line_number, line):
    """
    Ignores the frames of the asynchronous console.
    
    Parameters
    ----------
    file_name : `str`
        The frame's respective file's name.
    name : `str`
        The frame's respective function's name.
    line_number : `int`
        The line's index where the exception occurred.
    line : `str`
        The frame's respective stripped line.
    
    Returns
    -------
    should_show_frame : `bool`
        Whether the frame should be shown.
    """
    should_show_frame = True
    
    if file_name == __file__:
        if name == 'run_code_callback':
            if line == 'coroutine = function()':
                should_show_frame = False
        
        elif name == 'run_code':
            if line == 'result = future.sync_wrap().wait()':
                should_show_frame = False
            
            elif line == 'result = sync_wrapper.wait()':
                should_show_frame = False
    
    return should_show_frame


def run_code_callback(console, future, code):
    """
    Ensures the given `code` object on the event loop if awaitable. If not runs it instantly.
    
    Parameters
    ----------
    console : ``AsynchronousInteractiveConsole``
        The respective interactive console.
    future : ``Future``
        Future to set it's result.
    code : ``CodeType``
        The code to run.
    """
    function = FunctionType(code, console.local_variables)
    try:
        coroutine = function()
    except BaseException as err:
        future.set_exception_if_pending(err)
        return
    
    if not is_awaitable(coroutine):
        future.set_result_if_pending(coroutine)
        return
    
    try:
        task = console.loop.ensure_future(coroutine)
    except BaseException as err:
        future.set_exception_if_pending(err)
    else:
        console.task = task
        future.set_result_if_pending(None)


class AsynchronousInteractiveConsole:
    """
    Asynchronous interactive console.
    
    Attributes
    ----------
    _console_id : `int`
        The console's identifier.
    _file_name : `None`, `str`
        cached file name.
    _input_id : `int`
        Incremental value to update the file name of the input for each input chunk.
    banner : `str`
        banner shown when the console is started.
    editor_type : `type<EditorBase>`
        The specific editor type to use when inputting a expression.
    exit_message : `str`
        Message shown when the console is exited.
    highlighter : `None`, ``HighlightFormatterContext``
        Formatter storing highlighting details.
    history : ``History``
        History used for caching inputs.
    local_variables : `dict` of (`str`, `Any`) items
        Variables to run the compiled code with.
    loop : ``EventThread``
        The event loop on which the console runs it's code.
    stop_on_interruption : `bool`
        Whether the console should be stopped on keyboard interrupt.
    task : `None`, ``Future``
        Asynchronous task running on the event loop on what's result the console is waiting for.
    """
    _console_counter = 0
    
    __slots__ = (
        '_console_id', '_file_name', '_input_id', 'banner', 'editor_type', 'exit_message', 'highlighter', 'history',
        'local_variables', 'loop', 'stop_on_interruption', 'task',
    )
    
    def __new__(
        cls, *, banner=None, editor_type=None, event_loop=None, exit_message=None, highlighter=None,
        local_variables=None, stop_on_interruption=False
    ):
        """
        Creates a new async interactive console.
        
        Parameters
        ----------
        banner : `None`, `str` = `None`, Optional (Keyword only)
            Interactive console banner.
            
        editor_type : `type<EditorBase>`
            The specific editor type to use when inputting a expression.
            
        event_loop : `None`, ``EventThread`` = `None`, Optional (Keyword only)
                The event loop to run the executed code on.
        
        exit_message : `None`, `str` = `None`, Optional (Keyword only)
            Interactive console exit message.
        
        highlighter : `None`, ``HighlightFormatterContext`` = `None`, Optional (Keyword only)
            Formatter storing highlighting details.
        
        local_variables : `None`, `dict` of (`str`, `Any`) items = `None`, Optional (Keyword only)
            Initial local variables for the executed code inside of the console.
        
        stop_on_interruption : `bool` = `False`, Optional (Keyword only)
            Whether the console should be stopped on keyboard interrupt.
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        RuntimeError
            - If there are no event loops.
        """
        if event_loop is None:
            event_loop = get_event_loop()
        
        if editor_type is None:
            if can_use_advanced_editor():
                editor_type = EditorAdvanced
            
            else:
                editor_type = EditorSimple
        
        else:
            if (not isinstance(editor_type, type)) or (not issubclass(editor_type, EditorBase)):
                raise TypeError(
                    f'`editor_type` can be `None`, `{EditorBase.__name__}` subclass, got '
                    f'{editor_type.__class__.__name__}; {editor_type!r}.'
                )
        
        if exit_message is None:
            exit_message = create_exit_message()
        
        if highlighter is None:
            highlighter = get_default_trace_writer_highlighter()
        
        if banner is None:
            banner = create_banner(highlighter=highlighter)
        
        if local_variables is None:
            local_variables = {}
        
        console_id = cls._console_counter + 1
        cls._console_counter = console_id
        
        history = History()
        
        self = object.__new__(cls)
        
        self._console_id = console_id
        self._file_name = None
        self._input_id = 0
        self.banner = banner
        self.editor_type = editor_type
        self.exit_message = exit_message
        self.highlighter = highlighter
        self.history = history
        self.local_variables = local_variables
        self.loop = event_loop
        self.stop_on_interruption = stop_on_interruption
        self.task = None
        
        return self
    
    
    def run_code(self, compiled_code):
        """
        Runs the given code object on the respective event loop.
        
        Parameters
        ----------
        compiled_code : ``CodeType``
            The code to run.
        
        Returns
        -------
        result : `Any`
            The value returned by the ran code.
        
        Raises
        ------
        SystemExit
        """
        future = Future(self.loop)
        
        self.loop.call_soon_thread_safe(partial_func(run_code_callback, self, future, compiled_code))
        
        try:
            result = future.sync_wrap().wait()
            future = None
            
            task = self.task
            if (task is not None):
                
                sync_wrapper = task.sync_wrap()
                task = None
                
                try:
                    result = sync_wrapper.wait()
                except:
                    # Cancel the sync wrapper if exception occurs. This will stop warnings.
                    sync_wrapper.cancel()
                    raise
        
        except BaseException as err:
            if (future is not None):
                if (not future.is_done()):
                    future.cancel()
            
            task = self.task
            if (task is not None) and (not task.is_done()):
                task.cancel()
                task = None
            
            if isinstance(err, SystemExit):
                raise
            
            sys.stdout.write(
                ''.join(
                    render_exception_into(err, [], filter=_ignore_console_frames, highlighter=self.highlighter)
                )
            )
            
            if isinstance(err, KeyboardInterrupt):
                if self.stop_on_interruption:
                    sys.stdout.write('[Interrupt again to exit]\n')
        
        else:
            return result
        
        finally:
            self.task = None
            sys.stdout.flush()
    
    
    def raw_input(self, prefix):
        """
        Inputs from console with the given prompt.
        
        Parameters
        ----------
        prefix : `str`
            Prefix to start the line with.
        
        Returns
        -------
        input_value : `str`
        
        Raises
        ------
        BaseException
        """
        try:
            return input(prefix)
        except KeyboardInterrupt:
            if self.stop_on_interruption:
                sys.stdout.write('\n')
                sys.stdout.flush()
                raise SystemExit() from None
            
            raise
    
    
    def get_file_name(self):
        """
        Returns the file name of the current line.
        
        Returns
        -------
        file_name : `str`
        """
        file_name = self._file_name
        if (file_name is None):
            file_name = self._create_file_name()
            self._file_name = file_name
        
        return file_name
    
    
    def _create_file_name(self):
        """
        Creates the current file name of the console.
        
        Returns
        -------
        file_name : `str`
        """
        file_name_parts = ['<console']
        
        console_id = self._console_id
        if self._console_counter > console_id:
            file_name_parts.append('[')
            file_name_parts.append(str(console_id))
            file_name_parts.append(']')
        
        file_name_parts.append(' in[')
        file_name_parts.append(str(self._input_id))
        file_name_parts.append(']>')
        return ''.join(file_name_parts)
    
    
    def increment_input_counter(self):
        """
        Increments the console's input counter.
        """
        self._input_id += 1
        self._file_name = None
    
    
    def get_prefix_initial(self):
        """
        Returns the console's current initial prefix.
        
        Returns
        -------
        prefix_initial : `str`
        """
        marker = f'In [{self._input_id}]'
        prefix = ':'
        
        highlighter = self.highlighter
        if (highlighter is not None):
            marker = highlighter.highlight_as(
                marker, HIGHLIGHT_TOKEN_TYPES.TOKEN_TYPE_CONSOLE_MARKER_IN_INITIAL
            )
            
            prefix = highlighter.highlight_as(
                prefix, HIGHLIGHT_TOKEN_TYPES.TOKEN_TYPE_CONSOLE_MARKER_PREFIX_INITIAL
            )
        
        return f'{marker}{prefix} '
    
    
    def get_prefix_length(self):
        """
        Returns the console's current prefix's length.
        
        Returns
        -------
        prefix_length : `int`
        """
        input_id = self._input_id
        if input_id == 0:
            input_id_representation_length = 1
        else:
            input_id_representation_length = 1 + floor(log(input_id, 10))
        
        return input_id_representation_length + 7
    
    
    def get_prefix_continuous(self):
        """
        Returns the console's current continuous prefix.
        
        Returns
        -------
        continuous_prefix : `str`
        """
        marker_length = self.get_prefix_length() - 5
        
        indention = ' ' * marker_length
        marker = '...'
        prefix = ':'
        
        highlighter = self.highlighter
        if (highlighter is not None):
            marker = highlighter.highlight_as(
                marker, HIGHLIGHT_TOKEN_TYPES.TOKEN_TYPE_CONSOLE_MARKER_IN_CONTINUOUS
            )
            
            prefix = highlighter.highlight_as(
                prefix, HIGHLIGHT_TOKEN_TYPES.TOKEN_TYPE_CONSOLE_MARKER_PREFIX_CONTINUOUS
            )
        
        return f'{indention}{marker}{prefix} '
    
    
    def interact(self):
        """
        Emulates python interactive console.
        """
        sys.stdout.write(self.banner)
        sys.stdout.flush()
        editor = None
        
        while True:
            try:
                editor = self.editor_type(
                    None,
                    self.get_file_name(),
                    self.get_prefix_initial(),
                    self.get_prefix_continuous(),
                    self.get_prefix_length(),
                    self.highlighter,
                    self.history,
                )
                
                compiled_code = editor()
                if (compiled_code is not None):
                    self.run_code(compiled_code)
            
            except SyntaxError as err:
                self.history.maybe_add_buffer_of(self._input_id, editor)
                self.increment_input_counter()
                
                self.show_syntax_error(err, editor)
            
            except KeyboardInterrupt as err:
                self.history.maybe_add_buffer_of(self._input_id, editor)
                self.increment_input_counter()
                
                if self.stop_on_interruption:
                    if (editor is None) or editor.is_empty():
                        raise SystemExit() from err
                    
                    sys.stdout.write('\nKeyboardInterrupt [Interrupt again to exit]\n')
                    sys.stdout.flush()
                else:
                    sys.stdout.write('\nKeyboardInterrupt\n')
                    sys.stdout.flush()
            
            else:
                self.history.maybe_add_buffer_of(self._input_id, editor)
                self.increment_input_counter()
            
            finally:
                editor = None
        
        
        exit_message = self.exit_message
        sys.stdout.write(exit_message)
        if not exit_message.endswith('\n'):
            sys.stdout.write('\n')
        sys.stdout.flush()
    
    
    def show_syntax_error(self, syntax_error, editor):
        """
        Shows the given syntax error.
        
        Parameters
        ----------
        syntax_error : `SyntaxError`, `OverflowError`, `ValueError`
            The syntax error, or other derping to show.
        editor : `None`, ``EditorBase``
            The respective editor.
        """
        into = []
        
        if is_syntax_error(syntax_error):
            # message, (old_file_name, *additional_details) = syntax_error.args
            # syntax_error.args = (message, (self.get_file_name(), *additional_details))
            
            if (editor is not None):
                fixup_syntax_error_line_from_buffer(syntax_error, editor.get_buffer())
            
            _render_syntax_error_representation_into(syntax_error, into, self.highlighter)
            into.append('\n')
        else:
            render_exception_into(syntax_error, into, filter=_ignore_console_frames, highlighter=self.highlighter)
        
        sys.stdout.write(''.join(into))
        sys.stdout.flush()


def run_asynchronous_interactive_console(
    local_variables = None,
    *,
    banner = None,
    callback = None,
    exit_message = None,
    highlighter = None,
    stop_event_loop_when_done = True,
    stop_on_interruption = False,
):
    """
    Runs asynchronous interactive console.
    
    Parameters
    ----------
    local_variables : `None`, `dict` of (`str`, `Any`) items = `None`, Optional
        Local variables to start the console with.
    
    banner : `None`, `str` = `None`, Optional (Keyword only)
        Interactive console banner.
    
    callback : `None`, `FunctionType` = `None`, Optional (Keyword only)
        Callback to run when execution the console is closed.
        
        > This runs before the event loop is shut down!
    
    exit_message : `None`, `str` = `None`, Optional (Keyword only)
        Interactive console exit message.
    
    highlighter : `None`, ``HighlightFormatterContext`` = `None`, Optional (Keyword only)
        Formatter storing highlighting details.
    
    stop_event_loop_when_done : `bool` = `True`, Optional (Keyword only)
        Whether the event loop should be closed when the console is closed.
    
    stop_on_interruption : `bool` = `False`, Optional (Keyword only)
        Whether the console should be stopped on keyboard interrupt.
    """
    event_loop = get_or_create_event_loop()
    
    console = AsynchronousInteractiveConsole(
        banner = banner,
        event_loop = event_loop,
        exit_message = exit_message,
        highlighter = highlighter,
        local_variables = local_variables,
        stop_on_interruption = stop_on_interruption,
    )
    
    try:
        console.interact()
    finally:
        try:
            if (callback is not None):
                callback()
            
        finally:
            if stop_event_loop_when_done:
                warnings.filterwarnings(
                    'ignore',
                    message = r'^coroutine .* was never awaited$',
                    category = RuntimeWarning,
                )
                
                event_loop.stop()
